# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc

from chalk._gen.chalk.server.v1 import team_pb2 as chalk_dot_server_dot_v1_dot_team__pb2


class TeamServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetEnv = channel.unary_unary(
            "/chalk.server.v1.TeamService/GetEnv",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.GetEnvRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.GetEnvResponse.FromString,
        )
        self.GetEnvironments = channel.unary_unary(
            "/chalk.server.v1.TeamService/GetEnvironments",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.GetEnvironmentsRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.GetEnvironmentsResponse.FromString,
        )
        self.GetAgent = channel.unary_unary(
            "/chalk.server.v1.TeamService/GetAgent",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.GetAgentRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.GetAgentResponse.FromString,
        )
        self.GetDisplayAgent = channel.unary_unary(
            "/chalk.server.v1.TeamService/GetDisplayAgent",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.GetDisplayAgentRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.GetDisplayAgentResponse.FromString,
        )
        self.GetTeam = channel.unary_unary(
            "/chalk.server.v1.TeamService/GetTeam",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.GetTeamRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.GetTeamResponse.FromString,
        )
        self.CreateTeam = channel.unary_unary(
            "/chalk.server.v1.TeamService/CreateTeam",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.CreateTeamRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.CreateTeamResponse.FromString,
        )
        self.CreateProject = channel.unary_unary(
            "/chalk.server.v1.TeamService/CreateProject",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.CreateProjectRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.CreateProjectResponse.FromString,
        )
        self.CreateEnvironment = channel.unary_unary(
            "/chalk.server.v1.TeamService/CreateEnvironment",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.CreateEnvironmentRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.CreateEnvironmentResponse.FromString,
        )
        self.UpdateEnvironment = channel.unary_unary(
            "/chalk.server.v1.TeamService/UpdateEnvironment",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.UpdateEnvironmentRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.UpdateEnvironmentResponse.FromString,
        )
        self.GetAvailablePermissions = channel.unary_unary(
            "/chalk.server.v1.TeamService/GetAvailablePermissions",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.GetAvailablePermissionsRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.GetAvailablePermissionsResponse.FromString,
        )
        self.CreateServiceToken = channel.unary_unary(
            "/chalk.server.v1.TeamService/CreateServiceToken",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.CreateServiceTokenRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.CreateServiceTokenResponse.FromString,
        )
        self.DeleteServiceToken = channel.unary_unary(
            "/chalk.server.v1.TeamService/DeleteServiceToken",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.DeleteServiceTokenRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.DeleteServiceTokenResponse.FromString,
        )
        self.ListServiceTokens = channel.unary_unary(
            "/chalk.server.v1.TeamService/ListServiceTokens",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.ListServiceTokensRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.ListServiceTokensResponse.FromString,
        )
        self.UpdateServiceToken = channel.unary_unary(
            "/chalk.server.v1.TeamService/UpdateServiceToken",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.UpdateServiceTokenRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.UpdateServiceTokenResponse.FromString,
        )
        self.InviteTeamMember = channel.unary_unary(
            "/chalk.server.v1.TeamService/InviteTeamMember",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.InviteTeamMemberRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.InviteTeamMemberResponse.FromString,
        )
        self.ExpireTeamInvite = channel.unary_unary(
            "/chalk.server.v1.TeamService/ExpireTeamInvite",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.ExpireTeamInviteRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.ExpireTeamInviteResponse.FromString,
        )
        self.ListTeamInvites = channel.unary_unary(
            "/chalk.server.v1.TeamService/ListTeamInvites",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.ListTeamInvitesRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.ListTeamInvitesResponse.FromString,
        )
        self.UpsertFeaturePermissions = channel.unary_unary(
            "/chalk.server.v1.TeamService/UpsertFeaturePermissions",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.UpsertFeaturePermissionsRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.UpsertFeaturePermissionsResponse.FromString,
        )
        self.UpdateScimGroupSettings = channel.unary_unary(
            "/chalk.server.v1.TeamService/UpdateScimGroupSettings",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.UpdateScimGroupSettingsRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.UpdateScimGroupSettingsResponse.FromString,
        )
        self.GetTeamPermissions = channel.unary_unary(
            "/chalk.server.v1.TeamService/GetTeamPermissions",
            request_serializer=chalk_dot_server_dot_v1_dot_team__pb2.GetTeamPermissionsRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.GetTeamPermissionsResponse.FromString,
        )


class TeamServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetEnv(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetEnvironments(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetAgent(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetDisplayAgent(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetTeam(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CreateTeam(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CreateProject(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CreateEnvironment(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def UpdateEnvironment(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetAvailablePermissions(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CreateServiceToken(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def DeleteServiceToken(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListServiceTokens(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def UpdateServiceToken(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def InviteTeamMember(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ExpireTeamInvite(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListTeamInvites(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def UpsertFeaturePermissions(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def UpdateScimGroupSettings(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetTeamPermissions(self, request, context):
        """rpc SetEnvironmentVariables(SetEnvironmentVariablesRequest) returns
        (SetEnvironmentVariablesResponse) {
        option (chalk.auth.v1.permission) = PERMISSION_DEPLOY_CREATE;
        }

        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_TeamServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "GetEnv": grpc.unary_unary_rpc_method_handler(
            servicer.GetEnv,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.GetEnvRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.GetEnvResponse.SerializeToString,
        ),
        "GetEnvironments": grpc.unary_unary_rpc_method_handler(
            servicer.GetEnvironments,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.GetEnvironmentsRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.GetEnvironmentsResponse.SerializeToString,
        ),
        "GetAgent": grpc.unary_unary_rpc_method_handler(
            servicer.GetAgent,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.GetAgentRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.GetAgentResponse.SerializeToString,
        ),
        "GetDisplayAgent": grpc.unary_unary_rpc_method_handler(
            servicer.GetDisplayAgent,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.GetDisplayAgentRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.GetDisplayAgentResponse.SerializeToString,
        ),
        "GetTeam": grpc.unary_unary_rpc_method_handler(
            servicer.GetTeam,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.GetTeamRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.GetTeamResponse.SerializeToString,
        ),
        "CreateTeam": grpc.unary_unary_rpc_method_handler(
            servicer.CreateTeam,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.CreateTeamRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.CreateTeamResponse.SerializeToString,
        ),
        "CreateProject": grpc.unary_unary_rpc_method_handler(
            servicer.CreateProject,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.CreateProjectRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.CreateProjectResponse.SerializeToString,
        ),
        "CreateEnvironment": grpc.unary_unary_rpc_method_handler(
            servicer.CreateEnvironment,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.CreateEnvironmentRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.CreateEnvironmentResponse.SerializeToString,
        ),
        "UpdateEnvironment": grpc.unary_unary_rpc_method_handler(
            servicer.UpdateEnvironment,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.UpdateEnvironmentRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.UpdateEnvironmentResponse.SerializeToString,
        ),
        "GetAvailablePermissions": grpc.unary_unary_rpc_method_handler(
            servicer.GetAvailablePermissions,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.GetAvailablePermissionsRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.GetAvailablePermissionsResponse.SerializeToString,
        ),
        "CreateServiceToken": grpc.unary_unary_rpc_method_handler(
            servicer.CreateServiceToken,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.CreateServiceTokenRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.CreateServiceTokenResponse.SerializeToString,
        ),
        "DeleteServiceToken": grpc.unary_unary_rpc_method_handler(
            servicer.DeleteServiceToken,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.DeleteServiceTokenRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.DeleteServiceTokenResponse.SerializeToString,
        ),
        "ListServiceTokens": grpc.unary_unary_rpc_method_handler(
            servicer.ListServiceTokens,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.ListServiceTokensRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.ListServiceTokensResponse.SerializeToString,
        ),
        "UpdateServiceToken": grpc.unary_unary_rpc_method_handler(
            servicer.UpdateServiceToken,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.UpdateServiceTokenRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.UpdateServiceTokenResponse.SerializeToString,
        ),
        "InviteTeamMember": grpc.unary_unary_rpc_method_handler(
            servicer.InviteTeamMember,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.InviteTeamMemberRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.InviteTeamMemberResponse.SerializeToString,
        ),
        "ExpireTeamInvite": grpc.unary_unary_rpc_method_handler(
            servicer.ExpireTeamInvite,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.ExpireTeamInviteRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.ExpireTeamInviteResponse.SerializeToString,
        ),
        "ListTeamInvites": grpc.unary_unary_rpc_method_handler(
            servicer.ListTeamInvites,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.ListTeamInvitesRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.ListTeamInvitesResponse.SerializeToString,
        ),
        "UpsertFeaturePermissions": grpc.unary_unary_rpc_method_handler(
            servicer.UpsertFeaturePermissions,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.UpsertFeaturePermissionsRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.UpsertFeaturePermissionsResponse.SerializeToString,
        ),
        "UpdateScimGroupSettings": grpc.unary_unary_rpc_method_handler(
            servicer.UpdateScimGroupSettings,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.UpdateScimGroupSettingsRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.UpdateScimGroupSettingsResponse.SerializeToString,
        ),
        "GetTeamPermissions": grpc.unary_unary_rpc_method_handler(
            servicer.GetTeamPermissions,
            request_deserializer=chalk_dot_server_dot_v1_dot_team__pb2.GetTeamPermissionsRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_team__pb2.GetTeamPermissionsResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler("chalk.server.v1.TeamService", rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class TeamService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetEnv(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/GetEnv",
            chalk_dot_server_dot_v1_dot_team__pb2.GetEnvRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.GetEnvResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetEnvironments(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/GetEnvironments",
            chalk_dot_server_dot_v1_dot_team__pb2.GetEnvironmentsRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.GetEnvironmentsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetAgent(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/GetAgent",
            chalk_dot_server_dot_v1_dot_team__pb2.GetAgentRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.GetAgentResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetDisplayAgent(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/GetDisplayAgent",
            chalk_dot_server_dot_v1_dot_team__pb2.GetDisplayAgentRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.GetDisplayAgentResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetTeam(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/GetTeam",
            chalk_dot_server_dot_v1_dot_team__pb2.GetTeamRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.GetTeamResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CreateTeam(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/CreateTeam",
            chalk_dot_server_dot_v1_dot_team__pb2.CreateTeamRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.CreateTeamResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CreateProject(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/CreateProject",
            chalk_dot_server_dot_v1_dot_team__pb2.CreateProjectRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.CreateProjectResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CreateEnvironment(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/CreateEnvironment",
            chalk_dot_server_dot_v1_dot_team__pb2.CreateEnvironmentRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.CreateEnvironmentResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def UpdateEnvironment(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/UpdateEnvironment",
            chalk_dot_server_dot_v1_dot_team__pb2.UpdateEnvironmentRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.UpdateEnvironmentResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetAvailablePermissions(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/GetAvailablePermissions",
            chalk_dot_server_dot_v1_dot_team__pb2.GetAvailablePermissionsRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.GetAvailablePermissionsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CreateServiceToken(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/CreateServiceToken",
            chalk_dot_server_dot_v1_dot_team__pb2.CreateServiceTokenRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.CreateServiceTokenResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def DeleteServiceToken(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/DeleteServiceToken",
            chalk_dot_server_dot_v1_dot_team__pb2.DeleteServiceTokenRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.DeleteServiceTokenResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListServiceTokens(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/ListServiceTokens",
            chalk_dot_server_dot_v1_dot_team__pb2.ListServiceTokensRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.ListServiceTokensResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def UpdateServiceToken(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/UpdateServiceToken",
            chalk_dot_server_dot_v1_dot_team__pb2.UpdateServiceTokenRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.UpdateServiceTokenResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def InviteTeamMember(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/InviteTeamMember",
            chalk_dot_server_dot_v1_dot_team__pb2.InviteTeamMemberRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.InviteTeamMemberResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ExpireTeamInvite(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/ExpireTeamInvite",
            chalk_dot_server_dot_v1_dot_team__pb2.ExpireTeamInviteRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.ExpireTeamInviteResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListTeamInvites(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/ListTeamInvites",
            chalk_dot_server_dot_v1_dot_team__pb2.ListTeamInvitesRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.ListTeamInvitesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def UpsertFeaturePermissions(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/UpsertFeaturePermissions",
            chalk_dot_server_dot_v1_dot_team__pb2.UpsertFeaturePermissionsRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.UpsertFeaturePermissionsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def UpdateScimGroupSettings(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/UpdateScimGroupSettings",
            chalk_dot_server_dot_v1_dot_team__pb2.UpdateScimGroupSettingsRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.UpdateScimGroupSettingsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetTeamPermissions(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.TeamService/GetTeamPermissions",
            chalk_dot_server_dot_v1_dot_team__pb2.GetTeamPermissionsRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_team__pb2.GetTeamPermissionsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



from typing import List, Mapping

import pyarrow as pa

from chalk.features._encoding.pyarrow import pyarrow_to_polars
from chalk.utils.df_utils import pa_cast, pa_table_to_pl_df


def convert_hex_to_binary(table: pa.Table, cols_to_convert: List[str]) -> pa.Table:
    """
    Redshift returns binary data as a hex string.
    Ideally we'd add SQLAlchemy compiler hooks to do this conversion, but I'm not sure how to add this in the Chalk client w/o potentially clobbering other Redshift clients the customer may have.
    :param table: Table of data
    :param cols_to_convert: Columns to convert
    :return: Same table, but each column in cols_to_convert have been converterd from a hex string to binary
    """
    import polars as pl

    if len(cols_to_convert) == 0:
        return table
    pl_df = pa_table_to_pl_df(table)
    new_cols = [pl.col(colname).str.decode("hex").alias(colname) for colname in cols_to_convert]
    pl_df.with_columns(new_cols).to_arrow()


def json_parse_and_cast(tbl: pa.Table, schema: Mapping[str, pa.DataType]) -> pa.Table:
    import polars as pl

    """Json-parse any array and struct columns, do a case-insensitive column name mapping, and cast the table schema"""
    name_mapping = {k.upper(): k for k in schema}
    if len(schema) != len(name_mapping):
        raise ValueError(f"Column names must be case-insensitive. Case-insensitive names are {list(schema.keys())}")
    new_col_names: list[str] = []
    if len(name_mapping) != len(tbl.column_names):
        raise ValueError(
            f"Expected {len(name_mapping)} columns, got {len(tbl.column_names)} columns. Expected {list(name_mapping.keys())}; got {tbl.column_names}"
        )
    for x in tbl.column_names:
        cased_name = name_mapping.get(x.upper())
        if cased_name is None:
            raise ValueError(
                f"Did not expect to find column {x} (uppercased to {x.upper()}). Expected uppercase names {list(name_mapping.keys())}"
            )
        new_col_names.append(cased_name)

    tbl = tbl.rename_columns(new_col_names)
    # Snowflake returns arrays and structs as json strings. Parse them into their structured types
    json_cols_to_polars_dtype = {
        col_name: pyarrow_to_polars(pa_dtype, col_name)
        for (col_name, pa_dtype) in schema.items()
        if pa.types.is_struct(pa_dtype)
        or pa.types.is_list(pa_dtype)
        or pa.types.is_large_list(pa_dtype)
        or pa.types.is_fixed_size_list(pa_dtype)
    }
    if len(json_cols_to_polars_dtype) > 0:
        pl_df = pa_table_to_pl_df(tbl)
        pl_exprs: list[pl.Expr] = []
        for col_name, pl_dtype in json_cols_to_polars_dtype.items():
            expr = pl.col(col_name)
            if pl_df.schema[col_name] == pl.Binary():
                expr = expr.cast(pl.Utf8())
            expr = expr.str.json_extract(pl_dtype).alias(col_name)
            pl_exprs.append(expr)

        pl_df = pl_df.with_columns(pl_exprs)
        tbl = pl_df.to_arrow()
    tbl = pa_cast(tbl, pa.schema(schema))
    return tbl



from __future__ import annotations

from typing import List, Optional

from typing_extensions import Self

import chalk
from chalk.client.models import ChalkError

CHALK_TRACE_ID_KEY = "trace"
_CHALK_TRACE_ID_KEY = CHALK_TRACE_ID_KEY  # for backcompat
_CHALK_TRACE_ID_MESSAGE = "Trace ID:"
_CHALK_TRACE_ID_MISSING = "<trace ID missing>"


class ChalkBaseException(Exception):
    """The base type for Chalk exceptions.

    This exception makes error handling easier, as you can
    look only for this exception class.
    """

    errors: List[ChalkError]
    """The errors from executing a Chalk operation.

    These errors contain more detailed information about
    why the exception occurred.
    """

    def __init__(
        self, errors: Optional[List[ChalkError]] = None, trace_id: Optional[str] = None, detail: Optional[str] = None
    ):
        if errors is None:
            errors = []

        self.errors = errors
        self.trace_id = trace_id
        self.detail = detail

        super().__init__(self.full_message)

    @property
    def message(self) -> str:
        """A message describing the specific type of exception raised."""
        return f"Failed to execute Chalk operation (chalkpy=={chalk.__version__})"

    @property
    def full_message(self) -> str:
        """A message that describes the specific type of exception raised
        and contains the readable representation of each error in the
        errors attribute.

        Also includes the trace ID if one is available.
        """
        msg = self.message

        if self.errors:
            for e in self.errors[:10]:
                if e.display_primary_key:
                    msg += "\n\nPkey associated with error:\n\t"
                    if e.display_primary_key_fqn:
                        msg += f"{e.display_primary_key_fqn}: "
                    msg += f"{e.display_primary_key}"
                msg += f"\n\n{e.message}"
                if e.exception:
                    msg += f"\n{e.exception.stacktrace}"
                    msg += f"\n{e.exception.message}"

        if self.detail:
            msg += f"\n\nAdditional details: \n\t{self.detail}"

        if self.trace_id and _CHALK_TRACE_ID_MESSAGE not in msg:
            msg += f"\n\n{_CHALK_TRACE_ID_MESSAGE}: \n\t{self.trace_id}"

        return msg


class ChalkCustomException(ChalkBaseException):
    def __init__(
        self,
        message: str,
        errors: Optional[List[ChalkError]] = None,
        trace_id: Optional[str] = None,
        detail: Optional[str] = None,
    ):
        self._message = message
        super().__init__(errors=errors, trace_id=trace_id, detail=detail)

    @property
    def message(self) -> str:
        return self._message

    @classmethod
    def from_base(cls, base_exc: ChalkBaseException, message: str) -> Self:
        return cls(
            message=message,
            errors=base_exc.errors,
            trace_id=base_exc.trace_id,
            detail=base_exc.detail,
        )


class ChalkAuthException(ChalkBaseException):
    """Raised when constructing a `ChalkClient` without valid credentials.

    When this exception is raised, no explicit `client_id` and `client_secret`
    were provided, there was no `~/.chalk.yml` file with applicable credentials,
    and the environment variables `CHALK_CLIENT_ID` and `CHALK_CLIENT_SECRET`
    were not set.

    You may need to run `chalk login` from your command line, or check that your
    working directory is set to the root of your project.
    """

    @property
    def message(self):
        return (
            "Explicit `client_id` and `client_secret` are not provided, "
            "there is no `~/.chalk.yml` file with applicable credentials, "
            "and the environment variables `CHALK_CLIENT_ID` and "
            "`CHALK_CLIENT_SECRET` are not set. "
            "You may need to run `chalk login` from your command line, "
            "or check that your working directory is set to the root of "
            "your project."
        )


__all__ = [
    "ChalkBaseException",
    "ChalkCustomException",
    "ChalkAuthException",
]



from __future__ import annotations

from typing_extensions import final

from chalk.operators import StaticExpression, StaticOperator
from chalk.operators._column import ColumnExpression
from chalk.operators._literal import LiteralExpression


def get_name_for_expression(expr: StaticExpression) -> str:
    if isinstance(expr, LiteralExpression):
        if expr.name is None:
            raise ValueError("Literal expressions must have a name")
        return expr.name
    if isinstance(expr, ColumnExpression):
        return expr.final_name()
    raise ValueError("Expected a StaticExpression, got {expr}")


@final
class SelectOperator(StaticOperator):
    _chalk__operator_name = StaticOperator._chalk__operator_prefix + "select"

    def __init__(self, parent: StaticOperator, expressions: tuple[StaticExpression, ...]):
        self.column_names = tuple([get_name_for_expression(e) for e in expressions])
        self.expressions = expressions
        self.parent = parent
        super().__init__(parent, *expressions)



# pyright: reportPrivateUsage = false

from __future__ import annotations

import dataclasses
import warnings
from typing import Any, Tuple, Type, Union

import typing_extensions

from chalk.features.feature_field import Feature
from chalk.features.feature_wrapper import FeatureWrapper, unwrap_feature
from chalk.features.filter import Filter
from chalk.features.underscore import (
    SUPPORTED_UNDERSCORE_OPS_BINARY,
    SUPPORTED_UNDERSCORE_OPS_UNARY,
    DoubleUnderscore,
    Underscore,
    UnderscoreAttr,
    UnderscoreCall,
    UnderscoreCast,
    UnderscoreFunction,
    UnderscoreItem,
    UnderscoreRoot,
)
from chalk.utils.missing_dependency import missing_dependency_exception

if typing_extensions.TYPE_CHECKING:
    from chalk import Features

try:
    import polars as pl
except ModuleNotFoundError:
    pl = None


SUPPORTED_ARITHMETIC_OPS = {"+", "-", "*", "/", "//", "%", "**"}


def parse_underscore_in_context(exp: Underscore, context: Any, is_pydantic: bool = False) -> Any:
    """
    Parse a (potentially underscore) expression passed in under some "context".
    """
    parsed_exp = _parse_underscore_in_context(
        exp=exp,
        context=context,
        is_pydantic=is_pydantic,
    )
    assert not isinstance(parsed_exp, Underscore)
    return parsed_exp


def _parse_underscore_in_context(exp: Any, context: Any, is_pydantic: bool) -> Any:
    # Features of the dataframe are to be written as a dictionary of the fqn split up mapped to
    # the original features. The dictionary is represented immutably here.
    if not isinstance(exp, Underscore):
        # Recursive call hit non-underscore, deal with later
        return exp

    elif isinstance(exp, UnderscoreRoot):
        return context

    elif isinstance(exp, UnderscoreAttr):
        parent_context = _parse_underscore_in_context(exp=exp._chalk__parent, context=context, is_pydantic=is_pydantic)
        attr = exp._chalk__attr
        from chalk.features.dataframe import DataFrame

        if isinstance(parent_context, DataFrame) and is_pydantic:
            if attr not in parent_context._underlying.schema:
                warnings.warn(
                    f"Attribute {attr} not found in dataframe schema. Returning None. Found expression {exp}."
                )
                return None

            return attr
        else:
            return getattr(parent_context, attr)

    elif isinstance(exp, UnderscoreItem):
        parent_context = _parse_underscore_in_context(exp=exp._chalk__parent, context=context, is_pydantic=is_pydantic)
        key = exp._chalk__key
        return parent_context[key]

    elif isinstance(exp, UnderscoreCall):
        raise NotImplementedError(
            f"Calls on underscores in DataFrames is currently unsupported. Found expression {exp}"
        )

    elif isinstance(exp, UnderscoreFunction):
        if exp._chalk__function_name in SUPPORTED_UNDERSCORE_OPS_BINARY:
            if len(exp._chalk__args) != 2:
                raise ValueError(
                    f"Binary operation '{exp._chalk__function_name}' requires 2 operands; got {len(exp._chalk__args)} operands: {exp._chalk__args!r}"
                )
            left_val = exp._chalk__args[0]
            if isinstance(left_val, Underscore):
                left = _parse_underscore_in_context(exp=left_val, context=context, is_pydantic=is_pydantic)
            else:
                # The left value might be a literal, like `1` or `"::"` or `None`.
                left = left_val

            right_val = exp._chalk__args[1]
            if isinstance(right_val, Underscore):
                right = _parse_underscore_in_context(exp=right_val, context=context, is_pydantic=is_pydantic)
            else:
                # The right value might be a literal, like `1` or `"::"` or `None`.
                right = right_val

            if exp._chalk__function_name in SUPPORTED_ARITHMETIC_OPS:
                return _eval_arithmetic_expression(left, right, exp._chalk__function_name)
            else:
                return _eval_expression(left, right, exp._chalk__function_name)

        if exp._chalk__function_name in SUPPORTED_UNDERSCORE_OPS_UNARY:
            operand = _parse_underscore_in_context(exp=exp._chalk__args[0], context=context, is_pydantic=is_pydantic)
            return eval(f"{exp._chalk__function_name} operand", globals(), {"operand": operand})

    raise NotImplementedError(f"Unrecognized underscore expression {exp}")


def _unwrap_and_validate_features(left: FeatureWrapper, right: FeatureWrapper) -> Tuple[Feature, Feature]:
    f_left = unwrap_feature(left)
    f_right = unwrap_feature(right)

    if f_left.root_namespace != f_right.root_namespace:
        raise TypeError(
            f"{f_left} and {f_right} belong to different namespaces. Operations can only be performed on features of the same namespace."
        )

    return f_left, f_right


def _eval_expression(left: Union[FeatureWrapper, Filter], right: Any, op: str):
    try:
        if op == ">":
            return left > right
        elif op == "<":
            return left < right
        elif op == ">=":
            return left >= right
        elif op == "<=":
            return left <= right
        elif op == "==":
            return left == right
        elif op == "!=":
            return left != right
        elif op == "&":
            return left & right
        elif op == "|":
            return left | right
        elif op == "__getitem__":
            assert isinstance(left, FeatureWrapper)
            return left[right]
        elif op == "__getattr__":
            return getattr(left, right)
    except:
        raise NotImplementedError(
            f"Operation {op} not implemented for {type(left).__name__} and {type(right).__name__}"
        )


def _eval_arithmetic_expression(
    left: Union[FeatureWrapper, float, int],
    right: Union[FeatureWrapper, float, int],
    op: str,
):
    if pl is None:
        raise missing_dependency_exception("chalkpy[runtime]")

    if isinstance(left, FeatureWrapper) and isinstance(right, FeatureWrapper):
        # If both are features, ensure they are in the same namespace
        _unwrap_and_validate_features(left, right)

    if isinstance(left, FeatureWrapper):
        left_col = pl.col(str(left))
    else:
        left_col = pl.lit(left)

    if isinstance(right, FeatureWrapper):
        right_col = pl.col(str(right))
    else:
        right_col = pl.lit(right)

    if op == "+":
        return left_col + right_col
    elif op == "-":
        return left_col - right_col
    elif op == "*":
        return left_col * right_col
    elif op == "/":
        return left_col / right_col
    elif op == "//":
        return left_col // right_col
    elif op == "%":
        return left_col % right_col
    elif op == "**":
        return left_col**right_col

    raise NotImplementedError(f"{op} is not implemented")


@dataclasses.dataclass
class NamedUnderscoreExpr:
    """
    Associated a name with an underscore expression. Used to simulate 'creating' features in notebooks that can then be referred to in queries on the engine.
    """

    fqn: str
    short_name: str
    expr: Underscore

    def __str__(self):
        return f"{self.expr} as _.{self.short_name}"


@dataclasses.dataclass(frozen=True)
class NamedUnderscoreParseContext:
    current_namespace: str
    current_substitutions: tuple[str, ...]  # Each element is the fqn of a NamedUnderscoreExpression

    def with_substitution(self, new_expr: NamedUnderscoreExpr) -> NamedUnderscoreParseContext:
        return dataclasses.replace(self, current_substitutions=self.current_substitutions + (new_expr.fqn,))

    def with_namespace(self, new_namespace: str):
        return dataclasses.replace(self, current_namespace=new_namespace)


def process_named_underscore_expr(expr: NamedUnderscoreExpr):
    namespace = expr.fqn.split(".")[0]
    return _process_named_underscore_expr(
        expr=expr.expr, ctx=NamedUnderscoreParseContext(current_namespace=namespace, current_substitutions=tuple())
    )


def _process_named_underscore_expr(*, expr: Any, ctx: NamedUnderscoreParseContext):
    if isinstance(expr, NamedUnderscoreExpr):
        # Special case for if a user writes `FClass.some_new_feat` instead of `_.some_new_feat` in an expr
        # TODO check namespace here? `Rectangle.area = _.width * Square.height` is no good.
        return _process_named_underscore_expr(expr=expr.expr, ctx=ctx)

    if not isinstance(expr, Underscore):
        # If it's not an Underscore, return as-is
        return expr

    elif isinstance(expr, UnderscoreRoot):
        return expr
    elif isinstance(expr, UnderscoreAttr):
        parent_processed = _process_named_underscore_expr(expr=expr._chalk__parent, ctx=ctx)
        attr = expr._chalk__attr
        # For _.a.b.c, 'c' is in the namespace of "B" instead of the root namespace
        parent_namespace = get_namespace_of_underscore_expr(ctx.current_namespace, parent_processed)
        if parent_namespace is not None:
            if (named_expr := parent_namespace.__chalk_notebook_feature_expressions__.get(attr)) is not None:
                if named_expr.fqn in ctx.current_substitutions:
                    names = ctx.current_substitutions + (named_expr.fqn,)
                    cycle_str = " > ".join(names)
                    raise ValueError(
                        f"Circular reference when substituting locally-defined expressions with existing features: {cycle_str}"
                    )
                return _process_named_underscore_expr(
                    expr=named_expr.expr,
                    ctx=ctx.with_substitution(named_expr).with_namespace(parent_namespace.namespace),
                )
        return UnderscoreAttr(parent=parent_processed, attr=attr, expr_id=expr._chalk__expr_id)
    elif isinstance(expr, UnderscoreItem):
        # For _.a[_.b], b is a feature of the namespace of A
        parent_processed = _process_named_underscore_expr(expr=expr._chalk__parent, ctx=ctx)
        parent_namespace = get_namespace_of_underscore_expr(ctx.current_namespace, parent_processed)
        nested_ctx = ctx.with_namespace(parent_namespace.namespace) if parent_namespace is not None else ctx
        key_processed = tuple(_process_named_underscore_expr(expr=k, ctx=nested_ctx) for k in expr._chalk__key)
        return UnderscoreItem(parent=parent_processed, key=key_processed, expr_id=expr._chalk__expr_id)
    elif isinstance(expr, UnderscoreCall):
        parent_processed = _process_named_underscore_expr(expr=expr._chalk__parent, ctx=ctx)
        kwargs_processed = {k: _process_named_underscore_expr(expr=v, ctx=ctx) for k, v in expr._chalk__kwargs.items()}
        args_processed = [_process_named_underscore_expr(expr=a, ctx=ctx) for a in expr._chalk__args]
        return UnderscoreCall(
            parent=parent_processed, *args_processed, _chalk__expr_id=expr._chalk__expr_id, **kwargs_processed
        )
    elif isinstance(expr, UnderscoreFunction):
        kwargs_processed = {k: _process_named_underscore_expr(expr=v, ctx=ctx) for k, v in expr._chalk__kwargs.items()}
        args_processed = [_process_named_underscore_expr(expr=a, ctx=ctx) for a in expr._chalk__args]
        return UnderscoreFunction(
            expr._chalk__function_name, *args_processed, _chalk__expr_id=expr._chalk__expr_id, **kwargs_processed
        )
    elif isinstance(expr, UnderscoreCast):
        value_processed = _process_named_underscore_expr(expr=expr._chalk__value, ctx=ctx)
        return UnderscoreCast(to_type=expr._chalk__to_type, value=value_processed, expr_id=expr._chalk__expr_id)
    elif isinstance(expr, DoubleUnderscore):
        return expr

    raise NotImplementedError(f"Unrecognized underscore expression {expr}")


def get_namespace_of_underscore_expr(root_namespace: str, expr: Underscore) -> Type[Features] | None:
    if isinstance(expr, UnderscoreRoot):
        from chalk.features import FeatureSetBase

        return FeatureSetBase.registry.get(root_namespace)
    elif isinstance(expr, UnderscoreAttr):
        new_ns = get_namespace_of_underscore_expr(root_namespace, expr._chalk__parent)
        if new_ns is None:
            return None
        feat = None
        for x in new_ns.features:
            if x.name == expr._chalk__attr:
                feat = x
                break
        if feat is None:
            return None
        if feat.is_has_one:
            return feat.typ.as_features_cls()
        elif feat.is_has_many:
            df = feat.typ.as_dataframe()
            return df.references_feature_set if df is not None else None
    elif isinstance(expr, UnderscoreItem):
        # _.a[_.stuff] has the same namespace as _.a
        return get_namespace_of_underscore_expr(root_namespace, expr._chalk__parent)
    return None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/named_query.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2
from chalk._gen.chalk.graph.v1 import graph_pb2 as chalk_dot_graph_dot_v1_dot_graph__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n!chalk/server/v1/named_query.proto\x12\x0f\x63halk.server.v1\x1a\x1f\x63halk/auth/v1/permissions.proto\x1a\x1a\x63halk/graph/v1/graph.proto"@\n\x19GetAllNamedQueriesRequest\x12#\n\rdeployment_id\x18\x01 \x01(\tR\x0c\x64\x65ploymentId"0\n\x1aGetNamedQueryByNameRequest\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name"^\n\x1bGetNamedQueryByNameResponse\x12?\n\rnamed_queries\x18\x01 \x03(\x0b\x32\x1a.chalk.graph.v1.NamedQueryR\x0cnamedQueries"]\n\x1aGetAllNamedQueriesResponse\x12?\n\rnamed_queries\x18\x01 \x03(\x0b\x32\x1a.chalk.graph.v1.NamedQueryR\x0cnamedQueries"+\n)GetAllNamedQueriesActiveDeploymentRequest"m\n*GetAllNamedQueriesActiveDeploymentResponse\x12?\n\rnamed_queries\x18\x01 \x03(\x0b\x32\x1a.chalk.graph.v1.NamedQueryR\x0cnamedQueries2\xac\x03\n\x11NamedQueryService\x12u\n\x12GetAllNamedQueries\x12*.chalk.server.v1.GetAllNamedQueriesRequest\x1a+.chalk.server.v1.GetAllNamedQueriesResponse"\x06\x90\x02\x01\x80}\x0b\x12\xa5\x01\n"GetAllNamedQueriesActiveDeployment\x12:.chalk.server.v1.GetAllNamedQueriesActiveDeploymentRequest\x1a;.chalk.server.v1.GetAllNamedQueriesActiveDeploymentResponse"\x06\x90\x02\x01\x80}\x0b\x12x\n\x13GetNamedQueryByName\x12+.chalk.server.v1.GetNamedQueryByNameRequest\x1a,.chalk.server.v1.GetNamedQueryByNameResponse"\x06\x90\x02\x01\x80}\x0b\x42\x98\x01\n\x13\x63om.chalk.server.v1B\x0fNamedQueryProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.named_query_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\017NamedQueryProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_NAMEDQUERYSERVICE"].methods_by_name["GetAllNamedQueries"]._options = None
    _globals["_NAMEDQUERYSERVICE"].methods_by_name["GetAllNamedQueries"]._serialized_options = b"\220\002\001\200}\013"
    _globals["_NAMEDQUERYSERVICE"].methods_by_name["GetAllNamedQueriesActiveDeployment"]._options = None
    _globals["_NAMEDQUERYSERVICE"].methods_by_name[
        "GetAllNamedQueriesActiveDeployment"
    ]._serialized_options = b"\220\002\001\200}\013"
    _globals["_NAMEDQUERYSERVICE"].methods_by_name["GetNamedQueryByName"]._options = None
    _globals["_NAMEDQUERYSERVICE"].methods_by_name["GetNamedQueryByName"]._serialized_options = b"\220\002\001\200}\013"
    _globals["_GETALLNAMEDQUERIESREQUEST"]._serialized_start = 115
    _globals["_GETALLNAMEDQUERIESREQUEST"]._serialized_end = 179
    _globals["_GETNAMEDQUERYBYNAMEREQUEST"]._serialized_start = 181
    _globals["_GETNAMEDQUERYBYNAMEREQUEST"]._serialized_end = 229
    _globals["_GETNAMEDQUERYBYNAMERESPONSE"]._serialized_start = 231
    _globals["_GETNAMEDQUERYBYNAMERESPONSE"]._serialized_end = 325
    _globals["_GETALLNAMEDQUERIESRESPONSE"]._serialized_start = 327
    _globals["_GETALLNAMEDQUERIESRESPONSE"]._serialized_end = 420
    _globals["_GETALLNAMEDQUERIESACTIVEDEPLOYMENTREQUEST"]._serialized_start = 422
    _globals["_GETALLNAMEDQUERIESACTIVEDEPLOYMENTREQUEST"]._serialized_end = 465
    _globals["_GETALLNAMEDQUERIESACTIVEDEPLOYMENTRESPONSE"]._serialized_start = 467
    _globals["_GETALLNAMEDQUERIESACTIVEDEPLOYMENTRESPONSE"]._serialized_end = 576
    _globals["_NAMEDQUERYSERVICE"]._serialized_start = 579
    _globals["_NAMEDQUERYSERVICE"]._serialized_end = 1007
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import re
from datetime import timedelta
from typing import Collection, Iterable, List

_snake_sub_1 = re.compile("(.)([A-Z][a-z]+)")
_snake_sub_2 = re.compile("__([A-Z])")
_snake_sub_3 = re.compile("([a-z0-9])([A-Z])")


def to_snake_case(name: str):
    name = _snake_sub_1.sub(r"\1_\2", name)
    name = _snake_sub_2.sub(r"_\1", name)
    name = _snake_sub_3.sub(r"\1_\2", name)
    return name.lower()


def comma_whitespace_split(value: str):
    return re.split(r"\s*,\s*", value)


def comma_join(value: Iterable[str]):
    return ",".join(value)


_matching_re = re.compile(r"[^0-9a-z]")


def normalize_string_for_matching(name: str) -> str:
    """
    1. Case-insensitive
    2. Strip everything except alphanumeric, so e.g. `first_name` matches `firstname`
    """
    return _matching_re.sub("", name.lower())


def to_camel_case(snake_str: str) -> str:
    components = tuple(c for c in snake_str.split("_") if c != "")
    return components[0].lower() + "".join(x.title() for x in components[1:])


def to_title_case(snake_str: str):
    cameled = to_camel_case(snake_str)
    return cameled[0].upper() + cameled[1:] if len(cameled) > 0 else cameled


def oxford_comma_list(
    lst: Iterable[str],
    quoted: bool = False,
    quote_char: str = "'",
) -> str:
    lst_l = add_quotes(lst, quot_char=quote_char) if quoted else tuple(lst)
    len_l = len(lst_l)
    if len_l == 0:
        return ""
    elif len_l == 1:
        return lst_l[0]
    elif len_l == 2:
        return f"{lst_l[0]} and {lst_l[1]}"
    return ", ".join(lst_l[:-1]) + f", and {lst_l[-1]}"


def add_quotes(lst: Iterable[str], quot_char: str = "'") -> List[str]:
    return [f"{quot_char}{x}{quot_char}" for x in lst]


def double_quoted(s: str) -> str:
    return f'"{s}"'


def single_quoted(s: str) -> str:
    return f"'{s}'"


def double_quoted_list(lst: Iterable[str]) -> str:
    return ", ".join(map(double_quoted, lst))


def single_quoted_list(lst: Iterable[str]) -> str:
    return ", ".join(map(single_quoted, lst))


def to_bool(name: str, val: str | bool):
    if isinstance(val, bool):
        return val
    if val.lower() in ("yes", "y", "1", "true", "t"):
        return True
    if val.lower() in ("no", "no", "0", "false", "f"):
        return False
    raise ValueError(f"{name} must be 'yes' or 'no'")


def resolver_name(fqn: str) -> str:
    return fqn.split(".")[-1]


def matches_regex(string: str, regex: str) -> bool:
    pattern = re.compile(regex)
    return pattern.match(string) is not None


def in_regex_set(string: str | None, regex_set: Collection[str]) -> bool:
    if string is None:
        return False
    """Returns true if string is not None and in regex set"""
    return any(matches_regex(string, regex) for regex in regex_set)


def pluralize(word: str, items: Iterable[str]) -> str:
    items_l = sorted(items)
    if len(items_l) == 0:
        return f"no {word}s"
    elif len(items_l) == 1:
        return f"{word} '{items_l[0]}'"
    elif len(items_l) == 2:
        return f"{word}s '{items_l[0]}' and '{items_l[1]}'"

    content = ", '".join(items_l[:-1])
    return f"{word}s '{content}', and '{items_l[-1]}'"


def s(length: int):
    """
    Returns an "s" if 0-count or plural.
    Helps Chalk attain perfect grammar.

    :param length: Length of variable being described.
    :return: Empty string or "s".
    """
    return "" if length == 1 else "s"


def readable_duration(td: timedelta) -> str:
    seconds = int(td.total_seconds())
    milliseconds = int((td.microseconds / 1000000) * 1000)
    weeks, remainder = divmod(seconds, 604800)
    days, remainder = divmod(remainder, 86400)
    hours, remainder = divmod(remainder, 3600)
    minutes, seconds = divmod(remainder, 60)
    duration = []
    if weeks > 0:
        duration.append(f"{weeks} week{s(weeks)}")
    if days > 0:
        duration.append(f"{days} day{s(days)}")
    if hours > 0:
        duration.append(f"{hours} hour{s(hours)}")
    if minutes > 0:
        duration.append(f"{minutes} minute{s(minutes)}")
    if seconds > 0:
        duration.append(f"{seconds} second{s(seconds)}")
    if milliseconds > 0:
        duration.append(f"{milliseconds} millisecond{s(milliseconds)}")
    return " ".join(duration)



from typing import Optional

import chalk._gen.chalk.artifacts.v1.chart_pb2 as pb
from chalk._monitoring.Chart import Chart as _Chart
from chalk._monitoring.Chart import Trigger, _DatasetFeatureOperand, _Formula, _MultiSeriesOperand, _SingleSeriesOperand
from chalk._monitoring.charts_enums_codegen import (
    AlertSeverityKind,
    ChartLinkKind,
    ComparatorKind,
    FilterKind,
    GroupByKind,
    MetricFormulaKind,
    MetricKind,
    ThresholdPosition,
    WindowFunctionKind,
)
from chalk._monitoring.charts_series_base import MetricFilter, SeriesBase

_ENTITY_KIND_TO_PROTO = {
    ChartLinkKind.resolver: pb.ChartLinkKind.CHART_LINK_KIND_RESOLVER,
    ChartLinkKind.feature: pb.ChartLinkKind.CHART_LINK_KIND_FEATURE,
    ChartLinkKind.query: pb.ChartLinkKind.CHART_LINK_KIND_QUERY,
    ChartLinkKind.manual: pb.ChartLinkKind.CHART_LINK_KIND_MANUAL,
}

_METRIC_KIND_TO_PROTO = {
    MetricKind.FEATURE_REQUEST_COUNT: pb.MetricKind.METRIC_KIND_FEATURE_REQUEST_COUNT,
    MetricKind.FEATURE_LATENCY: pb.MetricKind.METRIC_KIND_FEATURE_LATENCY,
    MetricKind.FEATURE_STALENESS: pb.MetricKind.METRIC_KIND_FEATURE_STALENESS,
    MetricKind.FEATURE_VALUE: pb.MetricKind.METRIC_KIND_FEATURE_VALUE,
    MetricKind.FEATURE_WRITE: pb.MetricKind.METRIC_KIND_FEATURE_WRITE,
    MetricKind.FEATURE_NULL_RATIO: pb.MetricKind.METRIC_KIND_FEATURE_NULL_RATIO,
    MetricKind.RESOLVER_REQUEST_COUNT: pb.MetricKind.METRIC_KIND_RESOLVER_REQUEST_COUNT,
    MetricKind.RESOLVER_LATENCY: pb.MetricKind.METRIC_KIND_RESOLVER_LATENCY,
    MetricKind.RESOLVER_SUCCESS_RATIO: pb.MetricKind.METRIC_KIND_RESOLVER_SUCCESS_RATIO,
    MetricKind.QUERY_COUNT: pb.MetricKind.METRIC_KIND_QUERY_COUNT,
    MetricKind.QUERY_LATENCY: pb.MetricKind.METRIC_KIND_QUERY_LATENCY,
    MetricKind.QUERY_SUCCESS_RATIO: pb.MetricKind.METRIC_KIND_QUERY_SUCCESS_RATIO,
    MetricKind.BILLING_INFERENCE: pb.MetricKind.METRIC_KIND_BILLING_INFERENCE,
    MetricKind.BILLING_CRON: pb.MetricKind.METRIC_KIND_BILLING_CRON,
    MetricKind.BILLING_MIGRATION: pb.MetricKind.METRIC_KIND_BILLING_MIGRATION,
    MetricKind.CRON_COUNT: pb.MetricKind.METRIC_KIND_CRON_COUNT,
    MetricKind.CRON_LATENCY: pb.MetricKind.METRIC_KIND_CRON_LATENCY,
    MetricKind.STREAM_MESSAGES_PROCESSED: pb.MetricKind.METRIC_KIND_STREAM_MESSAGES_PROCESSED,
    MetricKind.STREAM_MESSAGE_LATENCY: pb.MetricKind.METRIC_KIND_STREAM_MESSAGE_LATENCY,
    MetricKind.STREAM_WINDOWS_PROCESSED: pb.MetricKind.METRIC_KIND_STREAM_WINDOWS_PROCESSED,
    MetricKind.STREAM_WINDOW_LATENCY: pb.MetricKind.METRIC_KIND_STREAM_WINDOW_LATENCY,
}

_FILTER_KIND_TO_PROTO = {
    FilterKind.FEATURE_STATUS: pb.FilterKind.FILTER_KIND_FEATURE_STATUS,
    FilterKind.FEATURE_NAME: pb.FilterKind.FILTER_KIND_FEATURE_NAME,
    FilterKind.FEATURE_TAG: pb.FilterKind.FILTER_KIND_FEATURE_TAG,
    FilterKind.RESOLVER_STATUS: pb.FilterKind.FILTER_KIND_RESOLVER_STATUS,
    FilterKind.RESOLVER_NAME: pb.FilterKind.FILTER_KIND_RESOLVER_NAME,
    FilterKind.RESOLVER_TAG: pb.FilterKind.FILTER_KIND_RESOLVER_TAG,
    FilterKind.CRON_STATUS: pb.FilterKind.FILTER_KIND_CRON_STATUS,
    FilterKind.MIGRATION_STATUS: pb.FilterKind.FILTER_KIND_MIGRATION_STATUS,
    FilterKind.ONLINE_OFFLINE: pb.FilterKind.FILTER_KIND_ONLINE_OFFLINE,
    FilterKind.CACHE_HIT: pb.FilterKind.FILTER_KIND_CACHE_HIT,
    FilterKind.OPERATION_ID: pb.FilterKind.FILTER_KIND_OPERATION_ID,
    FilterKind.QUERY_NAME: pb.FilterKind.FILTER_KIND_QUERY_NAME,
    FilterKind.QUERY_STATUS: pb.FilterKind.FILTER_KIND_QUERY_STATUS,
    FilterKind.IS_NULL: pb.FilterKind.FILTER_KIND_IS_NULL,
    FilterKind.USAGE_KIND: pb.FilterKind.FILTER_KIND_USAGE_KIND,
}

_COMPARATOR_KIND_TO_PROTO = {
    ComparatorKind.EQ: pb.ComparatorKind.COMPARATOR_KIND_EQ,
    ComparatorKind.NEQ: pb.ComparatorKind.COMPARATOR_KIND_NEQ,
    ComparatorKind.ONE_OF: pb.ComparatorKind.COMPARATOR_KIND_ONE_OF,
}

_WINDOW_FUNCTION_KIND_TO_PROTO = {
    WindowFunctionKind.COUNT: pb.WindowFunctionKind.WINDOW_FUNCTION_KIND_COUNT,
    WindowFunctionKind.MEAN: pb.WindowFunctionKind.WINDOW_FUNCTION_KIND_MEAN,
    WindowFunctionKind.SUM: pb.WindowFunctionKind.WINDOW_FUNCTION_KIND_SUM,
    WindowFunctionKind.MIN: pb.WindowFunctionKind.WINDOW_FUNCTION_KIND_MIN,
    WindowFunctionKind.MAX: pb.WindowFunctionKind.WINDOW_FUNCTION_KIND_MAX,
    WindowFunctionKind.PERCENTILE_99: pb.WindowFunctionKind.WINDOW_FUNCTION_KIND_PERCENTILE_99,
    WindowFunctionKind.PERCENTILE_95: pb.WindowFunctionKind.WINDOW_FUNCTION_KIND_PERCENTILE_95,
    WindowFunctionKind.PERCENTILE_75: pb.WindowFunctionKind.WINDOW_FUNCTION_KIND_PERCENTILE_75,
    WindowFunctionKind.PERCENTILE_50: pb.WindowFunctionKind.WINDOW_FUNCTION_KIND_PERCENTILE_50,
    WindowFunctionKind.PERCENTILE_25: pb.WindowFunctionKind.WINDOW_FUNCTION_KIND_PERCENTILE_25,
    WindowFunctionKind.PERCENTILE_5: pb.WindowFunctionKind.WINDOW_FUNCTION_KIND_PERCENTILE_5,
    WindowFunctionKind.ALL_PERCENTILES: pb.WindowFunctionKind.WINDOW_FUNCTION_KIND_ALL_PERCENTILES,
}

_GROUP_BY_KIND_TO_PROTO = {
    GroupByKind.FEATURE_STATUS: pb.GroupByKind.GROUP_BY_KIND_FEATURE_STATUS,
    GroupByKind.FEATURE_NAME: pb.GroupByKind.GROUP_BY_KIND_FEATURE_NAME,
    GroupByKind.IS_NULL: pb.GroupByKind.GROUP_BY_KIND_IS_NULL,
    GroupByKind.RESOLVER_STATUS: pb.GroupByKind.GROUP_BY_KIND_RESOLVER_STATUS,
    GroupByKind.RESOLVER_NAME: pb.GroupByKind.GROUP_BY_KIND_RESOLVER_NAME,
    GroupByKind.QUERY_STATUS: pb.GroupByKind.GROUP_BY_KIND_QUERY_STATUS,
    GroupByKind.QUERY_NAME: pb.GroupByKind.GROUP_BY_KIND_QUERY_NAME,
    GroupByKind.ONLINE_OFFLINE: pb.GroupByKind.GROUP_BY_KIND_ONLINE_OFFLINE,
    GroupByKind.CACHE_HIT: pb.GroupByKind.GROUP_BY_KIND_CACHE_HIT,
    GroupByKind.USAGE_KIND: pb.GroupByKind.GROUP_BY_KIND_USAGE_KIND,
}

_METRIC_FORMULA_KIND_TO_PROTO = {
    MetricFormulaKind.SUM: pb.MetricFormulaKind.METRIC_FORMULA_KIND_SUM,
    MetricFormulaKind.TOTAL_RATIO: pb.MetricFormulaKind.METRIC_FORMULA_KIND_TOTAL_RATIO,
    MetricFormulaKind.RATIO: pb.MetricFormulaKind.METRIC_FORMULA_KIND_RATIO,
    MetricFormulaKind.PRODUCT: pb.MetricFormulaKind.METRIC_FORMULA_KIND_PRODUCT,
    MetricFormulaKind.ABS: pb.MetricFormulaKind.METRIC_FORMULA_KIND_ABS,
    MetricFormulaKind.KS_STAT: pb.MetricFormulaKind.METRIC_FORMULA_KIND_KS_STAT,
    MetricFormulaKind.KS_TEST: pb.MetricFormulaKind.METRIC_FORMULA_KIND_KS_TEST,
    MetricFormulaKind.KS_THRESHOLD: pb.MetricFormulaKind.METRIC_FORMULA_KIND_KS_THRESHOLD,
    MetricFormulaKind.TIME_OFFSET: pb.MetricFormulaKind.METRIC_FORMULA_KIND_TIME_OFFSET,
}

_ALERT_SEVERITY_KIND_TO_PROTO = {
    AlertSeverityKind.CRITICAL: pb.AlertSeverityKind.ALERT_SEVERITY_KIND_CRITICAL,
    AlertSeverityKind.ERROR: pb.AlertSeverityKind.ALERT_SEVERITY_KIND_ERROR,
    AlertSeverityKind.WARNING: pb.AlertSeverityKind.ALERT_SEVERITY_KIND_WARNING,
    AlertSeverityKind.INFO: pb.AlertSeverityKind.ALERT_SEVERITY_KIND_INFO,
}

_THRESHOLD_POSITION_KIND_TO_PROTO = {
    ThresholdPosition.ABOVE: pb.ThresholdKind.THRESHOLD_KIND_ABOVE,
    ThresholdPosition.BELOW: pb.ThresholdKind.THRESHOLD_KIND_BELOW,
    ThresholdPosition.GREATER_EQUAL: pb.ThresholdKind.THRESHOLD_KIND_GREATER_EQUAL,
    ThresholdPosition.LESS_EQUAL: pb.ThresholdKind.THRESHOLD_KIND_LESS_EQUAL,
    ThresholdPosition.EQUAL: pb.ThresholdKind.THRESHOLD_KIND_EQUAL,
    ThresholdPosition.NOT_EQUAL: pb.ThresholdKind.THRESHOLD_KIND_NOT_EQUAL,
}


def _convert_metric_kind(metric_kind: Optional[MetricKind], chart_name: str, series_name: str) -> pb.MetricKind:
    if metric_kind is None:
        raise ValueError(
            f"Chart '{chart_name}' has a series '{series_name}' with no metric. "
            f"'metric' is a required value for Series instances"
        )
    res = _METRIC_KIND_TO_PROTO.get(metric_kind)
    if res is None:
        raise ValueError(f"Chart '{chart_name}' has an invalid metric '{metric_kind}'")
    return res


def _convert_filter_kind(filter_kind: FilterKind) -> pb.FilterKind:
    res = _FILTER_KIND_TO_PROTO.get(filter_kind)
    if res is None:
        raise ValueError(f"Invalid filter kind '{filter_kind}'")
    return res


def _convert_comparator_kind(comparator_kind: ComparatorKind) -> pb.ComparatorKind:
    res = _COMPARATOR_KIND_TO_PROTO.get(comparator_kind)
    if res is None:
        raise ValueError(f"Invalid comparator kind '{comparator_kind}'")
    return res


def _convert_window_function_kind(
    window_function: Optional[WindowFunctionKind],
) -> Optional["pb.WindowFunctionKind"]:
    if window_function is None:
        return None
    res = _WINDOW_FUNCTION_KIND_TO_PROTO.get(window_function)
    if res is None:
        raise ValueError(f"Invalid window function kind '{window_function}'")
    return res


def _convert_group_by_kind(group_by: GroupByKind) -> pb.GroupByKind:
    res = _GROUP_BY_KIND_TO_PROTO.get(group_by)
    if res is None:
        raise ValueError(f"Invalid group by kind '{group_by}'")
    return res


def _convert_alert_severity_kind(
    alert_severity: Optional[AlertSeverityKind], chart_name: str, trigger_name: str
) -> pb.AlertSeverityKind:
    if not alert_severity:
        raise ValueError(
            f"Chart '{chart_name}' has a trigger '{trigger_name} with no severity level. "
            f"'severity' is a required value for Trigger instances"
        )
    res = _ALERT_SEVERITY_KIND_TO_PROTO.get(alert_severity)
    if res is None:
        raise ValueError(f"Invalid alert severity kind '{alert_severity}'")
    return res


def _convert_threshold_position(
    threshold_position: Optional[ThresholdPosition], chart_name: str, trigger_name: str
) -> pb.ThresholdKind:
    if not threshold_position:
        raise ValueError(
            f"Chart '{chart_name}' has a trigger '{trigger_name} with no threshold position. "
            f"'threshold_position' is a required value for Trigger instances"
        )
    res = _THRESHOLD_POSITION_KIND_TO_PROTO.get(threshold_position)
    if res is None:
        raise ValueError(f"Invalid threshold position kind '{threshold_position}'")
    return res


def _convert_metric_formula_kind(
    kind: Optional[MetricFormulaKind], chart_name: str, formula_name: str
) -> pb.MetricFormulaKind:
    if not kind:
        raise ValueError(
            f"Chart '{chart_name}' has a formula '{formula_name}' with no operation kind. "
            f"'kind' is a required value for Formula instances"
        )
    res = _METRIC_FORMULA_KIND_TO_PROTO.get(kind)
    if res is None:
        raise ValueError(f"Invalid metric formula kind '{kind}'")
    return res


def _convert_filter(filter: MetricFilter) -> pb.MetricFilter:
    return pb.MetricFilter(
        kind=_convert_filter_kind(filter.kind),
        comparator=_convert_comparator_kind(filter.comparator),
    )


def _convert_series(series: SeriesBase, chart_name: str) -> pb.MetricConfigSeries:
    for validation_fn in series._validations:
        validation_fn()
    return pb.MetricConfigSeries(
        metric=_convert_metric_kind(series._metric, chart_name, series._name),
        filters=[_convert_filter(series_filter) for series_filter in series._filters],
        name=series._name if series._name else series._default_name,
        window_function=_convert_window_function_kind(series._window_function),
        group_by=series._group_by,
    )


def _convert_formula(formula: _Formula, chart_name: str) -> pb.MetricFormula:
    operands = formula._operands
    if not operands:
        raise ValueError(
            f"Chart '{chart_name}' has a formula '{formula._name}' with no operands. "
            f"'operands' is a required value for Formula instances"
        )
    single_series = None
    multi_series = None
    dataset_feature = None
    if isinstance(operands, _SingleSeriesOperand):
        single_series = operands.operand
    elif isinstance(operands, _MultiSeriesOperand):
        multi_series = operands.operands
    elif isinstance(operands, _DatasetFeatureOperand):
        dataset_feature = pb.DatasetFeatureOperand(dataset=operands.dataset, feature=operands.feature)
    else:
        raise ValueError(
            f"Chart '{chart_name}' has a formula {formula._name}' with an invalid value for operand '{operands}'"
        )
    return pb.MetricFormula(
        kind=_convert_metric_formula_kind(formula._kind, chart_name, formula._name),
        single_series_operands=single_series,
        multi_series_operands=multi_series,
        dataset_feature_operands=dataset_feature,
    )


def _convert_trigger(trigger: Trigger, chart_name: str) -> pb.AlertTrigger:
    if not trigger._name:
        raise ValueError(
            f"Chart {chart_name} has a trigger with no name. " f"'name' is a required value for Trigger instances"
        )
    if trigger._threshold_value is None:
        raise ValueError(
            "Chart {chart_name} has a trigger with no threshold value. "
            "'threshold_value' is a required value for Trigger instances"
        )
    return pb.AlertTrigger(
        name=trigger._name,
        severity=_convert_alert_severity_kind(trigger._severity, chart_name, trigger._name),
        threshold_position=_convert_threshold_position(trigger._threshold_position, chart_name, trigger._name),
        threshold_value=trigger._threshold_value,
        series_name=trigger._series_name,
        channel_name=trigger._channel_name,
        description=trigger._description,
    )


def convert_chart(chart: _Chart) -> pb.Chart:
    if not chart._window_period:
        raise ValueError(
            f"Chart {chart._name} has no window period. " f"window_period' is a required value for Chart instances"
        )
    entity_kind = _ENTITY_KIND_TO_PROTO.get(chart._entity_kind)
    if entity_kind is None:
        raise ValueError(f"Chart {chart._name} has an invalid entity_kind '{chart._entity_kind}'")
    config = pb.MetricConfig(
        name=chart._name,
        window_period=chart._window_period,
        series=[_convert_series(series, chart._name) for series in chart._series],
        formulas=[_convert_formula(formula, chart._name) for formula in chart._formulas],
        trigger=_convert_trigger(chart._trigger, chart._name) if chart._trigger else None,
    )
    return pb.Chart(
        id=str(hash(chart)),
        config=config,
        entity_kind=entity_kind,
        entity_id=chart._entity_id,
    )



from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from chalk._gen.chalk.server.v1 import scheduled_query_run_pb2 as _scheduled_query_run_pb2
from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class ManualTriggerScheduledQueryRequest(_message.Message):
    __slots__ = ("cron_query_id", "planner_options", "incremental_resolvers", "max_samples", "env_overrides")
    class PlannerOptionsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _struct_pb2.Value
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[_Union[_struct_pb2.Value, _Mapping]] = ...
        ) -> None: ...

    class EnvOverridesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    CRON_QUERY_ID_FIELD_NUMBER: _ClassVar[int]
    PLANNER_OPTIONS_FIELD_NUMBER: _ClassVar[int]
    INCREMENTAL_RESOLVERS_FIELD_NUMBER: _ClassVar[int]
    MAX_SAMPLES_FIELD_NUMBER: _ClassVar[int]
    ENV_OVERRIDES_FIELD_NUMBER: _ClassVar[int]
    cron_query_id: int
    planner_options: _containers.MessageMap[str, _struct_pb2.Value]
    incremental_resolvers: _containers.RepeatedScalarFieldContainer[str]
    max_samples: int
    env_overrides: _containers.ScalarMap[str, str]
    def __init__(
        self,
        cron_query_id: _Optional[int] = ...,
        planner_options: _Optional[_Mapping[str, _struct_pb2.Value]] = ...,
        incremental_resolvers: _Optional[_Iterable[str]] = ...,
        max_samples: _Optional[int] = ...,
        env_overrides: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class ManualTriggerScheduledQueryResponse(_message.Message):
    __slots__ = ("scheduled_query_run",)
    SCHEDULED_QUERY_RUN_FIELD_NUMBER: _ClassVar[int]
    scheduled_query_run: _scheduled_query_run_pb2.ScheduledQueryRun
    def __init__(
        self, scheduled_query_run: _Optional[_Union[_scheduled_query_run_pb2.ScheduledQueryRun, _Mapping]] = ...
    ) -> None: ...



import collections
import os
import re
from datetime import timedelta
from pathlib import Path
from typing import Collection, List, Optional, Sequence, Tuple

import chalk._gen.chalk.artifacts.v1.export_pb2 as export_pb
from chalk._gen.chalk.artifacts.v1 import chart_pb2 as chart_pb
from chalk._gen.chalk.artifacts.v1.cdc_pb2 import CDCSource, CDCTableReference
from chalk._gen.chalk.artifacts.v1.cron_query_pb2 import CronQuery, RecomputeSettings
from chalk._gen.chalk.common.v1 import chalk_error_pb2
from chalk._gen.chalk.lsp.v1 import lsp_pb2
from chalk._lsp.error_builder import LSPErrorBuilder
from chalk._monitoring.Chart import Chart as _Chart
from chalk._monitoring.proto_conversion import convert_chart
from chalk._version import __version__
from chalk.client import ChalkError, ChalkException, ErrorCode, ErrorCodeCategory
from chalk.client.serialization.protos import ChalkErrorConverter
from chalk.config.project_config import ProjectSettings, load_project_config
from chalk.features import FeatureSetBase
from chalk.features.resolver import RESOLVER_REGISTRY
from chalk.importer import CHALK_IMPORTER, import_all_files
from chalk.parsed._proto.lsp import convert_lsp_gql_to_proto
from chalk.parsed._proto.utils import (
    build_failed_import,
    convert_failed_import_to_gql,
    convert_failed_import_to_proto,
    datetime_to_proto_timestamp,
)
from chalk.parsed._proto.validation import validate_artifacts
from chalk.parsed.to_proto import ToProtoConverter
from chalk.parsed.user_types_to_json import get_lsp_gql
from chalk.queries.named_query import NAMED_QUERY_REGISTRY
from chalk.queries.scheduled_query import CRON_QUERY_REGISTRY
from chalk.sql._internal.sql_source import BaseSQLSource, TableIngestMixIn
from chalk.sql._internal.sql_source_group import SQLSourceGroup
from chalk.streams import StreamSource
from chalk.utils.duration import timedelta_to_duration


def project_settings_to_proto(config: ProjectSettings) -> export_pb.ProjectSettings:
    def read_packages(filename: str) -> Optional[List[str]]:
        reqs: List[str] = []
        try:
            with open(filename) as f:
                for r in f.readlines():
                    cleaned = re.sub("#.*", "", r).rstrip("\n").strip()
                    if cleaned != "":
                        reqs.append(cleaned)
            return reqs

        except OSError:
            return None

    environments: List[export_pb.EnvironmentSettings] = []
    envs = config.environments
    if envs is not None:
        environments = [
            export_pb.EnvironmentSettings(
                id=env_id,
                runtime=e.runtime,
                requirements=e.requirements,
                dockerfile=e.dockerfile,
                requires_packages=(
                    None
                    if e.requirements is None
                    else read_packages(
                        os.path.join(
                            os.path.dirname(config.local_path),
                            e.requirements,
                        )
                    )
                ),
                platform_version=e.platform_version,
            )
            for env_id, e in envs.items()
        ]

    validation: Optional[export_pb.ValidationSettings] = None
    if config.validation:
        feature_settings: Optional[export_pb.FeatureSettings] = None
        if config.validation.feature:
            metadata_settings = (
                [export_pb.MetadataSettings(name=m.name, missing=m.missing) for m in config.validation.feature.metadata]
                if config.validation.feature.metadata
                else None
            )
            feature_settings = export_pb.FeatureSettings(metadata=metadata_settings)

        resolver_settings: Optional[export_pb.ResolverSettings] = None
        if config.validation.resolver:
            metadata_settings = (
                [
                    export_pb.MetadataSettings(name=m.name, missing=m.missing)
                    for m in config.validation.resolver.metadata
                ]
                if config.validation.resolver.metadata
                else None
            )
            resolver_settings = export_pb.ResolverSettings(metadata=metadata_settings)

        validation = export_pb.ValidationSettings(
            feature=feature_settings,
            resolver=resolver_settings,
        )

    return export_pb.ProjectSettings(
        project=config.project,
        environments=environments,
        validation=validation,
    )


def import_files_then_export_from_registry(
    directory: Optional[str], file_allowlist: Optional[List[str]]
) -> export_pb.Export:
    """
    Packaged into a function just so that the caller can try-except this.
    """
    if directory is not None:
        os.chdir(directory)
    failed_imports_gql = import_all_files(
        file_allowlist=file_allowlist,
        project_root=None if directory is None else Path(directory),
    )
    export = export_from_registry()
    export.failed.extend([convert_failed_import_to_proto(f) for f in failed_imports_gql])

    return export


def get_lsp_proto_or_error(
    additional_failed_import_protos: Sequence[export_pb.FailedImport],
) -> Tuple[lsp_pb2.LSP, Optional[ChalkException]]:
    """
    :param additional_failed_import_protos:  Supplementary failed imports to add as lsp diagnostics
    :return: A tuple [LSP, error | None]. If generating the LSP threw an exception, return the partial LSP and a serialized form of the error for reporting.
    """
    lsp = get_lsp_gql()
    exc: Optional[ChalkException] = None
    if LSPErrorBuilder.lsp:
        failed_gql = [convert_failed_import_to_gql(f) for f in additional_failed_import_protos]
        try:
            # Modified lsp.diagnostics in-place to add info.
            # if we throw an exception due to bad state, we'll still have the errors we've collected so far
            CHALK_IMPORTER.supplement_diagnostics(failed_imports=failed_gql, diagnostics=lsp.diagnostics)
        except Exception as e:
            if len(lsp.diagnostics) == 0 and len(failed_gql) == 0:
                exc = ChalkException.from_exception(e)
    return convert_lsp_gql_to_proto(lsp), exc


def get_lsp_proto(
    additional_failed_import_protos: Sequence[export_pb.FailedImport],
) -> lsp_pb2.LSP:
    lsp, _ = get_lsp_proto_or_error(additional_failed_import_protos)
    return lsp


def export_from_registry() -> export_pb.Export:
    """
    This is separate from trying to `import_all_files` so that
    we can use this function in places where import is already
    done, like the engine.
    """
    failed_protos: List[export_pb.FailedImport] = []

    graph_res = ToProtoConverter.convert_graph(
        features_registry=FeatureSetBase.registry,
        resolver_registry=RESOLVER_REGISTRY.get_all_resolvers(),
        sql_source_registry=BaseSQLSource.registry,
        sql_source_group_registry=SQLSourceGroup.registry,
        stream_source_registry=StreamSource.registry,
        named_query_registry=NAMED_QUERY_REGISTRY,
    )

    crons: List[CronQuery] = []
    for cron in CRON_QUERY_REGISTRY.values():
        if not isinstance(cron.recompute_features, (Collection, bool)):  # pyright: ignore[reportUnnecessaryIsInstance]
            failed_protos.append(
                build_failed_import(
                    TypeError(f"Invalid `recompute_features` type '{type(cron.recompute_features).__name__}'"),
                    f"cron '{cron.name}'",
                )
            )
            continue
        crons.append(
            CronQuery(
                name=cron.name,
                cron=timedelta_to_duration(cron.cron) if isinstance(cron.cron, timedelta) else cron.cron,
                output=[str(f) for f in cron.output],
                max_samples=cron.max_samples,
                recompute=RecomputeSettings(
                    feature_fqns=list(cron.recompute_features)
                    if isinstance(cron.recompute_features, Collection)
                    else None,
                    all_features=cron.recompute_features  # pyright: ignore[reportArgumentType]
                    if isinstance(cron.recompute_features, bool)
                    else None,
                ),
                lower_bound=datetime_to_proto_timestamp(cron.lower_bound)
                if cron.lower_bound is not None
                else cron.lower_bound,
                upper_bound=datetime_to_proto_timestamp(cron.upper_bound)
                if cron.upper_bound is not None
                else cron.upper_bound,
                tags=cron.tags,
                required_resolver_tags=cron.required_resolver_tags,
                store_online=cron.store_online,
                store_offline=cron.store_offline,
                file_name=cron.filename,
                resource_group=cron.resource_group,
                planner_options=cron.planner_options,
            )
        )

    charts: List[chart_pb.Chart] = []
    for chart in _Chart.registry:
        try:
            charts.append(convert_chart(chart))
        except Exception as e:
            failed_protos.append(build_failed_import(e, f"chart ' {chart.name}'"))

    integration_name_to_tables = collections.defaultdict(list)
    for source in BaseSQLSource.registry:
        if isinstance(source, TableIngestMixIn):
            for schema_dot_table, preferences in source.ingested_tables.items():
                if preferences.cdc is True:
                    assert isinstance(source, BaseSQLSource)
                    parts = schema_dot_table.split(".")
                    if len(parts) != 2:
                        build_failed_import(
                            f"Expected {{schema}}.{{table}}, got {schema_dot_table}", f"Database source '{source.name}'"
                        )
                        continue
                    integration_name_to_tables[source.name].append(CDCTableReference(schema=parts[0], name=parts[1]))
    cdc_sources = [
        CDCSource(
            integration_name=integration_name,
            tables=tables,
        )
        for integration_name, tables in integration_name_to_tables.items()
    ]

    config = load_project_config()
    if config is not None:
        config = project_settings_to_proto(config)

    # Has side effect of populating LSPErrorBuilder
    logs = validate_artifacts(graph_res, config)

    # Get proto of currently logged LSP diagnostics
    lsp, exc = get_lsp_proto_or_error(failed_protos)
    errors: List[chalk_error_pb2.ChalkError] = []
    if exc is not None:
        err = ChalkError.create(
            code=ErrorCode.INTERNAL_SERVER_ERROR,
            category=ErrorCodeCategory.NETWORK,
            message=f"Failed to export features: {exc.message}",
            exception=exc,
        )
        errors.append(ChalkErrorConverter.chalk_error_encode(err))

    return export_pb.Export(
        graph=graph_res,
        crons=crons,
        charts=charts,
        chalkpy=export_pb.ChalkpyInfo(version=__version__),
        cdc_sources=cdc_sources,
        config=config,
        logs=logs,
        lsp=lsp,
        failed=failed_protos,
        conversion_errors=errors,
    )



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.engine.v2.feature_values_chart_pb2 import (
    GetFeatureValuesTimeSeriesChartRequest,
    GetFeatureValuesTimeSeriesChartResponse,
)
from chalk._gen.chalk.engine.v2.feature_values_pb2 import (
    GetFeatureValuesRequest,
    GetFeatureValuesResponse,
)
from chalk._gen.chalk.engine.v2.query_log_pb2 import (
    GetQueryLogEntriesRequest,
    GetQueryLogEntriesResponse,
)
from chalk._gen.chalk.engine.v2.query_values_pb2 import (
    GetQueryValuesRequest,
    GetQueryValuesResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class OfflineStoreServiceStub:
    """This service exposes endpoints for dealing with the offline store. It should never depend on the python graph.
    v2 introduces two breaking changes:
    Uses messages from engine.v2 instead of common.v1 (common is not meant for engine-specific messages)
    Removes certain endpoints added and had to be immediately deprecated due to deprecation
    """

    def __init__(self, channel: Channel) -> None: ...
    GetQueryLogEntries: UnaryUnaryMultiCallable[
        GetQueryLogEntriesRequest,
        GetQueryLogEntriesResponse,
    ]
    GetQueryValues: UnaryUnaryMultiCallable[
        GetQueryValuesRequest,
        GetQueryValuesResponse,
    ]
    GetFeatureValuesTimeSeriesChart: UnaryUnaryMultiCallable[
        GetFeatureValuesTimeSeriesChartRequest,
        GetFeatureValuesTimeSeriesChartResponse,
    ]
    GetFeatureValues: UnaryUnaryMultiCallable[
        GetFeatureValuesRequest,
        GetFeatureValuesResponse,
    ]

class OfflineStoreServiceServicer(metaclass=ABCMeta):
    """This service exposes endpoints for dealing with the offline store. It should never depend on the python graph.
    v2 introduces two breaking changes:
    Uses messages from engine.v2 instead of common.v1 (common is not meant for engine-specific messages)
    Removes certain endpoints added and had to be immediately deprecated due to deprecation
    """

    @abstractmethod
    def GetQueryLogEntries(
        self,
        request: GetQueryLogEntriesRequest,
        context: ServicerContext,
    ) -> GetQueryLogEntriesResponse: ...
    @abstractmethod
    def GetQueryValues(
        self,
        request: GetQueryValuesRequest,
        context: ServicerContext,
    ) -> GetQueryValuesResponse: ...
    @abstractmethod
    def GetFeatureValuesTimeSeriesChart(
        self,
        request: GetFeatureValuesTimeSeriesChartRequest,
        context: ServicerContext,
    ) -> GetFeatureValuesTimeSeriesChartResponse: ...
    @abstractmethod
    def GetFeatureValues(
        self,
        request: GetFeatureValuesRequest,
        context: ServicerContext,
    ) -> GetFeatureValuesResponse: ...

def add_OfflineStoreServiceServicer_to_server(servicer: OfflineStoreServiceServicer, server: Server) -> None: ...



from __future__ import annotations

import types
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Mapping, Optional, Sequence, Union
from uuid import UUID

from chalk import DataFrame
from chalk.client.models import (
    BulkOnlineQueryResponse,
    FeatureReference,
    FeatureStatisticsResponse,
    OfflineQueryInputUri,
    OnlineQuery,
    OnlineQueryContext,
    ResourceRequests,
    UploadFeaturesResponse,
    WhoAmIResponse,
)
from chalk.client.response import Dataset, OnlineQueryResult
from chalk.features._encoding.json import FeatureEncodingOptions
from chalk.features.tag import BranchId, DeploymentId, EnvironmentId
from chalk.prompts import Prompt

if TYPE_CHECKING:
    import pandas as pd
    import polars as pl

    QueryInput = Mapping[FeatureReference, Any] | pd.DataFrame | pl.DataFrame | DataFrame | str


class AsyncChalkClient:
    """The `AsyncChalkClient` is an asynchronous Python interface for interacting with Chalk.

    You can use it to query data, trigger resolver runs, gather offline data, and more, and all calls are asynchronous.
    """

    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        environment: Optional[EnvironmentId] = None,
        api_server: Optional[str] = None,
        query_server: Optional[str] = None,
        branch: Optional[BranchId] = None,
        preview_deployment_id: Optional[DeploymentId] = None,
        additional_headers: Optional[Mapping[str, str]] = None,
        default_job_timeout: float | timedelta | None = None,
        default_request_timeout: float | timedelta | None = None,
        executor: Optional[ThreadPoolExecutor] = None,
        pool_maxsize: Optional[int] = None,
    ):
        """Create an `AsyncChalkClient` with the given credentials.

        Parameters
        ----------
        client_id
            The client ID to use to authenticate. Can either be a
            service token id or a user token id.
        client_secret
            The client secret to use to authenticate. Can either be a
            service token secret or a user token secret.
        environment
            The ID or name of the environment to use for this client.
            Not necessary if your `client_id` and `client_secret`
            are for a service token scoped to a single environment.
            If not present, the client will use the environment variable
            `CHALK_ENVIRONMENT`.
        api_server
            The API server to use for this client. Required if you are
            using a Chalk Dedicated deployment. If not present, the client
            will check for the presence of the environment variable
            `CHALK_API_SERVER`, and use that if found.
        query_server
            The query server to use for this client. Required if you are
            using a standalone Chalk query engine deployment. If not present,
            the client will default to the value of `api_server`.
        branch
            If specified, Chalk will route all requests from this client
            instance to the relevant branch. Some methods allow you to
            override this instance-level branch configuration by passing
            in a `branch` argument.
        preview_deployment_id
            If specified, Chalk will route all requests from this client
            instance to the relevant preview deployment.
        additional_headers
            A map of additional HTTP headers to pass with each request.
        default_job_timeout
            The default wait timeout, in seconds, to wait for long-running jobs to complete
            when accessing query results.
            Jobs will not time out if this timeout elapses. For no timeout, set to `None`.
            The default timeout is 10 minutes.
        default_request_timeout:
            The default wait timeout, in seconds, to wait for network requests to complete.
            If not specified, the default is no timeout.
        executor
            A custom `ThreadPoolExecutor` to use for running asynchronous tasks.
            If not provided, will use Chalk's default thread pool.
        pool_maxsize
            The maximum number of connections the HTTP session is configured to handle.
            If not provided, will use the default pool size.

        Raises
        ------
        ChalkAuthException
            If `client_id` or `client_secret` are not provided, there
            is no `~/.chalk.yml` file with applicable credentials,
            and the environment variables `CHALK_CLIENT_ID` and
            `CHALK_CLIENT_SECRET` are not set.
        """
        super().__init__()
        ...

    def __new__(cls, *args: Any, **kwargs: Any):
        from chalk.client.client_async_impl import AsyncChalkClientImpl

        return AsyncChalkClientImpl(*args, **kwargs)

    async def whoami(self) -> WhoAmIResponse:
        """Checks the identity of your client.

        Useful as a sanity test of your configuration.

        Returns
        -------
        WhoAmIResponse
            The identity of your client.

        Examples
        --------
        >>> from chalk.client import AsyncChalkClient
        >>> await AsyncChalkClient().whoami()
        WhoAmIResponse(user="...", environment_id='...', team_id='...')
        """
        ...

    async def query(
        self,
        input: Mapping[FeatureReference, Any] | Any,
        output: Sequence[FeatureReference] = (),
        now: datetime | None = None,
        staleness: Mapping[FeatureReference, str] | None = None,
        environment: EnvironmentId | None = None,
        tags: list[str] | None = None,
        preview_deployment_id: str | None = None,
        branch: BranchId | None = ...,
        correlation_id: str | None = None,
        query_name: str | None = None,
        query_name_version: str | None = None,
        include_meta: bool = False,
        meta: Mapping[str, str] | None = None,
        explain: bool = False,
        store_plan_stages: bool = False,
        encoding_options: FeatureEncodingOptions | None = None,
        required_resolver_tags: list[str] | None = None,
        planner_options: Mapping[str, Union[str, int, bool]] | None = None,
        request_timeout: Optional[float] = None,
        connect_timeout: Optional[float] = None,
        headers: Mapping[str, str] | None = None,
        query_context: Mapping[str, Union[str, int, float, bool, None]] | str | None = None,
    ) -> OnlineQueryResult:
        """Compute features values using online resolvers.
        See https://docs.chalk.ai/docs/query-basics for more information.

        Parameters
        ----------
        input
            The features for which there are known values, mapped to those values.
            For example, `{User.id: 1234}`. Features can also be expressed as snakecased strings,
            e.g. `{"user.id": 1234}`
        output
            Outputs are the features that you'd like to compute from the inputs.
            For example, `[User.age, User.name, User.email]`.

            If an empty sequence, the output will be set to all features on the namespace
            of the query. For example, if you pass as input `{"user.id": 1234}`, then the query
            is defined on the `User` namespace, and all features on the `User` namespace
            (excluding has-one and has-many relationships) will be used as outputs.
        staleness
            Maximum staleness overrides for any output features or intermediate features.
            See https://docs.chalk.ai/docs/query-caching for more information.
        environment
            The environment under which to run the resolvers.
            API tokens can be scoped to an environment.
            If no environment is specified in the query,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        tags
            The tags used to scope the resolvers.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        required_resolver_tags
            If specified, *all* required_resolver_tags must be present on a resolver for it to be
            considered eligible to execute.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        branch
            If specified, Chalk will route your request to the relevant branch.
        preview_deployment_id
            If specified, Chalk will route your request to the relevant preview deployment.
        query_name
            The semantic name for the query you're making, for example, `"loan_application_model"`.
            Typically, each query that you make from your application should have a name.
            Chalk will present metrics and dashboard functionality grouped by 'query_name'.
        include_meta
            Returns metadata about the query execution under `OnlineQueryResult.meta`.
            This could make the query slightly slower.
            For more information, see https://docs.chalk.ai/docs/query-basics.
        explain
            Log the query execution plan. Requests using `explain=True` will be slower
            than requests using `explain=False`.

            If `True`, 'include_meta' will be set to `True` as well.
        store_plan_stages
            If `True`, the output of each of the query plan stages will be stored.
            This option dramatically impacts the performance of the query,
            so it should only be used for debugging.
        correlation_id
            You can specify a correlation ID to be used in logs and web interfaces.
            This should be globally unique, i.e. a `uuid` or similar. Logs generated
            during the execution of your query will be tagged with this correlation id.
        now
            The time at which to evaluate the query. If not specified, the current time will be used.
            This parameter is complex in the context of online_query since the online store
            only stores the most recent value of an entity's features. If `now` is in the past,
            it is extremely likely that `None` will be returned for cache-only features.

            This parameter is primarily provided to support:
                - controlling the time window for aggregations over cached has-many relationships
                - controlling the time window for aggregations over has-many relationships loaded from an
                  external database
            If you are trying to perform an exploratory analysis of past feature values, prefer `offline_query`.
        query_context
            An immutable context that can be accessed from Python resolvers.
            This context wraps a JSON-compatible dictionary or JSON string with type restrictions.
            See https://docs.chalk.ai/api-docs#ChalkContext for more information.

        Other Parameters
        ----------------
        meta
            Arbitrary `key:value` pairs to associate with a query.
        planner_options
            Dictionary of additional options to pass to the Chalk query engine.
            Values may be provided as part of conversations with Chalk support
            to enable or disable specific functionality.
        request_timeout
            Float value indicating number of seconds that the request should wait before timing out
            at the network level. May not cancel resources on the server processing the query.
        connect_timeout
            Float value indicating number of seconds to wait for establishing a connection.
            This is separate from request_timeout and controls only the connection phase.
        headers
            Additional headers to provide with the request.

        Returns
        -------
        Awaitable[OnlineQueryResult]
            Coroutine that returns a wrapper around the output features and any query metadata,
            plus errors encountered while running the resolvers.

        Examples
        --------
        >>> from chalk.client import AsyncChalkClient
        >>> result = await AsyncChalkClient().query(
        ...     input={
        ...         User.name: "Katherine Johnson"
        ...     },
        ...     output=[User.fico_score],
        ...     staleness={User.fico_score: "10m"},
        ... )
        >>> result.get_feature_value(User.fico_score)
        """
        ...

    async def multi_query(
        self,
        queries: list[OnlineQuery],
        environment: EnvironmentId | None = None,
        preview_deployment_id: str | None = None,
        branch: BranchId | None = ...,
        correlation_id: str | None = None,
        query_name: str | None = None,
        query_name_version: str | None = None,
        query_context: Mapping[str, Union[str, int, float, bool, None]] | str | None = None,
        meta: Mapping[str, str] | None = None,
        use_feather: bool | None = True,
        compression: str | None = "uncompressed",
    ) -> BulkOnlineQueryResponse:
        """
        Execute multiple queries (represented by `queries=` argument) in a single request. This is useful if the
        queries are "rooted" in different `@features` classes -- i.e. if you want to load features for `User` and
        `Merchant` and there is no natural relationship object which is related to both of these classes, `multi_query`
        allows you to submit two independent queries.

        Returns a BulkOnlineQueryResponse, which is functionally a list of query results. Each of these result
        can be accessed by index. Individual results can be further checked for errors and converted
        to pandas or polars DataFrames.

        In contrast, `query_bulk` executes a single query with multiple inputs/outputs.

        Parameters
        ----------
        queries
            A list of the OnlineQueries you'd like to execute.
        environment
            The environment under which to run the resolvers.
            API tokens can be scoped to an environment.
            If no environment is specified in the query,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        branch
            If specified, Chalk will route your request to the relevant branch.
        preview_deployment_id
            If specified, Chalk will route your request to the
            relevant preview deployment.

        Other Parameters
        ----------------
        query_name
            The name for class of query you're making, for example, `"loan_application_model"`.
        query_context
            An immutable context that can be accessed from Python resolvers.
            This context wraps a JSON-compatible dictionary or JSON string with type restrictions.
            See https://docs.chalk.ai/api-docs#ChalkContext for more information.
        correlation_id
            A globally unique ID for the query, used alongside logs and
            available in web interfaces.
        meta
            Arbitrary `key:value` pairs to associate with a query.
        query_context
            An immutable context that can be accessed from Python resolvers.
            This context wraps a JSON-compatible dictionary or JSON string with type restrictions.
            See https://docs.chalk.ai/api-docs#ChalkContext for more information.
        compression
            Which compression scheme to use pyarrow. Options are: {"zstd", "lz4", "uncompressed"}.

        Returns
        -------
        Awaitable[BulkOnlineQueryResponse]
            Coroutine that returns object containing results: list[BulkOnlineQueryResult], where each result contains
            dataframes of the results of each query or any errors.

        Examples
        --------
        >>> from chalk.client import AsyncChalkClient, OnlineQuery
        >>> queries =[
        ...     OnlineQuery(input={User.name: ['Katherine Johnson'], output=[User.fico_score]}),
        ...     OnlineQuery(input={Merchant.name: ['Myrrh Chant'], output=['Merchant.address']}),
        ...     OnlineQuery(input={NonFeature.wrong: ['Wrong!'], output=['NonFeature.wrong']}),
        ... ]
        >>> result = await AsyncChalkClient().multi_query(
        ...     queries=queries,
        ... )
        >>> result[0].get_feature_value(User.fico_score)
        >>> queries_with_errors = [q for q, r in zip(queries, result) if r.errors is not None]
        """
        ...

    async def query_bulk(
        self,
        input: Mapping[FeatureReference, Sequence[Any]],
        output: Sequence[FeatureReference] = (),
        now: Sequence[datetime] | None = None,
        staleness: Mapping[FeatureReference, str] | None = None,
        context: OnlineQueryContext | None = None,  # Deprecated.
        environment: EnvironmentId | None = None,
        store_plan_stages: bool = False,
        tags: list[str] | None = None,
        required_resolver_tags: list[str] | None = None,
        preview_deployment_id: str | None = None,
        branch: BranchId | None = ...,
        correlation_id: str | None = None,
        query_name: str | None = None,
        query_name_version: str | None = None,
        query_context: Mapping[str, Union[str, int, float, bool, None]] | str | None = None,
        meta: Mapping[str, str] | None = None,
        explain: bool = False,
        request_timeout: Optional[float] = None,
        headers: Mapping[str, str] | None = None,
    ) -> BulkOnlineQueryResponse:
        """Compute features values for many rows of inputs using online resolvers.
        See https://docs.chalk.ai/docs/query-basics for more information on online query.

        This method is similar to `query`, except it takes in `list` of inputs, and produces one
        output per row of inputs.

        This method is appropriate if you want to fetch the same set of features for many different
        input primary keys.

        This method contrasts with `multi_query`, which executes multiple fully independent queries.

        This endpoint is not available in all environments.

        Parameters
        ----------
        input
            The features for which there are known values, mapped to a list
            of the values.
        output
            Outputs are the features that you'd like to compute from the inputs.
        staleness
            Maximum staleness overrides for any output features or intermediate features.
            See https://docs.chalk.ai/docs/query-caching for more information.
        environment
            The environment under which to run the resolvers.
            API tokens can be scoped to an environment.
            If no environment is specified in the query,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        tags
            The tags used to scope the resolvers.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        branch
            If specified, Chalk will route your request to the relevant branch.
        preview_deployment_id
            If specified, Chalk will route your request to the
            relevant preview deployment.
        now
            The time at which to evaluate the query. If not specified, the current time will be used.
            The length of this list must be the same as the length of the values in `input`.

        Other Parameters
        ----------------
        query_name
            The name for class of query you're making, for example, `"loan_application_model"`.
        query_context
            An immutable context that can be accessed from Python resolvers.
            This context wraps a JSON-compatible dictionary or JSON string with type restrictions.
            See https://docs.chalk.ai/api-docs#ChalkContext for more information.
        correlation_id
            A globally unique ID for the query, used alongside logs and
            available in web interfaces.
        meta
            Arbitrary `key:value` pairs to associate with a query.
        context
            Deprecated in favor of `environment` and `tags`.
        request_timeout
            Float value indicating number of seconds that the request should wait before timing out
            at the network level. May not cancel resources on the server processing the query
        explain
            Log the query execution plan. Requests using `explain=True` will be slower
            than requests using `explain=False`.
        headers
            Additional headers to provide with the request

        Returns
        -------
        Awaitable[BulkOnlineQueryResponse]
            Coroutine of a list[BulkOnlineQueryResult], where each result contains dataframes of the
            results of each query.

        Examples
        --------
        >>> from chalk.client import AsyncChalkClient
        >>> await AsyncChalkClient().query_bulk(
        ...     input={User.name: ["Katherine Johnson", "Eleanor Roosevelt"]},
        ...     output=[User.fico_score],
        ...     staleness={User.fico_score: "10m"},
        ... )
        """
        ...

    async def offline_query(
        self,
        input: Union[QueryInput, OfflineQueryInputUri] | None = None,
        input_times: Sequence[datetime] | datetime | None = None,
        output: Sequence[FeatureReference] = (),
        required_output: Sequence[FeatureReference] = (),
        environment: EnvironmentId | None = None,
        dataset_name: str | None = None,
        branch: BranchId | None = ...,
        correlation_id: str | None = None,
        query_context: Mapping[str, Union[str, int, float, bool, None]] | str | None = None,
        max_samples: int | None = None,
        wait: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | ellipsis | None = ...,
        recompute_features: bool | list[FeatureReference] = False,
        sample_features: list[FeatureReference] | None = None,
        lower_bound: datetime | timedelta | str | None = None,
        upper_bound: datetime | timedelta | str | None = None,
        store_plan_stages: bool = False,
        explain: bool = False,
        tags: list[str] | None = None,
        required_resolver_tags: list[str] | None = None,
        planner_options: Mapping[str, Union[str, int, bool]] | None = None,
        spine_sql_query: str | None = None,
        resources: ResourceRequests | None = None,
        run_asynchronously: bool = False,
        store_online: bool = False,
        store_offline: bool = False,
        num_shards: int | None = None,
        num_workers: int | None = None,
        completion_deadline: timedelta | None = None,
        max_retries: int | None = None,
    ) -> Dataset:
        """Compute feature values from the offline store or by running offline/online resolvers.
        See `Dataset` for more information.

        Parameters
        ----------
        input
            The features for which there are known values.
            It can be a mapping of features to a list of values for each
            feature, or an existing `DataFrame`.
            Each element in the `DataFrame` or list of values represents
            an observation in line with the timestamp in `input_times`.
        input_times
            A list of the times of the observations from `input`.
        output
            The features that you'd like to sample, if they exist.
            If an output feature was never computed for a sample (row) in
            the resulting `DataFrame`, its value will be `None`.
        recompute_features
            Used to control whether resolvers are allowed to run in order to compute feature values.

            If True, all output features will be recomputed by resolvers.
            If False, all output features will be sampled from the offline store.
            If a list, all output features in recompute_features will be recomputed,
            and all other output features will be sampled from the offline store.
        sample_features
            A list of features that will always be sampled, and thus always excluded from recompute.
            Should not overlap with any features used in `recompute_features` argument.
        environment
            The environment under which to run the resolvers.
            API tokens can be scoped to an environment.
            If no environment is specified in the query,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        dataset_name
            A unique name that if provided will be used to generate and
            save a `Dataset` constructed from the list of features computed
            from the inputs.
        max_samples
            The maximum number of samples to include in the `DataFrame`.
            If not specified, all samples will be returned.
        branch
            If specified, Chalk will route your request to the relevant branch.
            If None, Chalk will route your request to a non-branch deployment.
            If not specified, Chalk will use the current client's branch info.
        correlation_id
            You can specify a correlation ID to be used in logs and web interfaces.
            This should be globally unique, i.e. a `uuid` or similar. Logs generated
            during the execution of your query will be tagged with this correlation id.
        query_context
            An immutable context that can be accessed from Python resolvers.
            This context wraps a JSON-compatible dictionary or JSON string with type restrictions.
            See https://docs.chalk.ai/api-docs#ChalkContext for more information.
        wait
            Whether to wait for job completion
        show_progress
            If True, progress bars will be shown while the query is running.
            Primarily intended for use in a Jupyter-like notebook environment.
            This flag will also be propagated to the methods of the resulting
            `Dataset`.
        timeout:
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.
        lower_bound
            If specified, the query will only be run on data observed after this timestamp.
            Accepts strings in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
        upper_bound
            If specified, the query will only be run on data observed before this timestamp.
            Accepts strings in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
        store_plan_stages
            If True, the output of each of the query plan stages will be stored
            in S3/GCS. This will dramatically impact the performance of the query,
            so it should only be used for debugging.
            These files will be visible in the web dashboard's query detail view, and
            can be downloaded in full by clicking on a plan node in the query plan visualizer.
        tags
            The tags used to scope the resolvers.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        required_resolver_tags
            If specified, *all* required_resolver_tags must be present on a resolver for it to be
            considered eligible to execute.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        store_online
            If True, the output of the query will be stored in the online store.
        store_offline
            If True, the output of the query will be stored in the offline store.
        num_shards
            If specified, the query will be run asynchronously, splitting the input across `num_shards` shards.
        num_workers
            If specified, the query will be run asynchronously across a maximum `num_workers` pod workers at any time.
            This parameter is useful if you have a large number of shards and would like to limit the number of pods running at once.

        Other Parameters
        ----------------
        required_output
            The features that you'd like to sample and must exist
            in each resulting row. Rows where a `required_output`
            was never stored in the offline store will be skipped.
            This differs from specifying the feature in `output`,
            where instead the row would be included, but the feature
            value would be `None`.

        Returns
        -------
        Dataset
            A Chalk `Dataset`.

        Examples
        --------
        >>> from chalk.client import AsyncChalkClient
        >>> uids = [1, 2, 3, 4]
        >>> at = datetime.now(tz=timezone.utc)
        >>> dataset = await AsyncChalkClient().offline_query(
        ...     input={
        ...         User.id: uids,
        ...     },
        ...     input_times=[at] * len(uids),
        ...     output=[
        ...         User.id,
        ...         User.fullname,
        ...         User.email,
        ...         User.name_email_match_score,
        ...     ],
        ...     dataset_name='my_dataset'
        ... )
        >>> df = dataset.get_data_as_pandas()
        """
        ...

    async def create_dataset(
        self,
        input: QueryInput,
        dataset_name: str | None = None,
        environment: EnvironmentId | None = None,
        branch: BranchId | None = ...,
        wait: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | ellipsis | None = ...,
    ) -> Dataset:
        """Create a Chalk `Dataset`.

        The `Dataset` wraps a lazily-loading Chalk `DataFrame` that enables us to analyze
        our data without loading all of it directly into memory.
        See https://docs.chalk.ai/docs/query-offline for more information.

        Parameters
        ----------
        input
            The features for which there are known values.
            It can be a mapping of features to a list of values for each
            feature, or an existing `DataFrame`.
        dataset_name
            A unique name that if provided will be used to generate and
            save a `Dataset` constructed from the inputs.
        environment
            The environment under which to execute the request.
            API tokens can be scoped to an environment.
            If no environment is specified in the request,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        wait
            Whether to wait for job completion.
        show_progress
            If True, progress bars will be shown while the query is running.
            Primarily intended for use in a Jupyter-like notebook environment.
            This flag will also be propagated to the methods of the resulting
            `Dataset`.
        timeout:
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        Dataset
            A Chalk `Dataset`.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> uids = [1, 2, 3, 4]
        >>> names = ['a', 'b', 'c', 'd']
        >>> dataset = ChalkClient().create_dataset(
        ...     input={
        ...         User.id: uids,
        ...         User.name: names,
        ...     },
        ...     dataset_name='my_dataset'
        ... )
        >>> df = dataset.get_data_as_pandas()
        """
        ...

    async def prompt_evaluation(
        self,
        prompts: list[Prompt | str],
        dataset_id: str | None = None,
        dataset_name: str | None = None,
        revision_id: str | None = None,
        reference_output: FeatureReference | None = None,
        evaluators: list[str] | None = None,
        meta: Mapping[str, str] | None = None,
        input: QueryInput | None = None,
        input_times: Sequence[datetime] | datetime | None = None,
        output: Sequence[FeatureReference] = (),
        required_output: Sequence[FeatureReference] = (),
        environment: EnvironmentId | None = None,
        branch: BranchId | None = ...,
        correlation_id: str | None = None,
        query_context: Mapping[str, Union[str, int, float, bool, None]] | str | None = None,
        max_samples: int | None = None,
        wait: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | ellipsis | None = ...,
        recompute_features: bool | list[FeatureReference] = False,
        sample_features: list[FeatureReference] | None = None,
        lower_bound: datetime | timedelta | str | None = None,
        upper_bound: datetime | timedelta | str | None = None,
        store_plan_stages: bool = False,
        explain: bool = False,
        tags: list[str] | None = None,
        required_resolver_tags: list[str] | None = None,
        planner_options: Mapping[str, Union[str, int, bool]] | None = None,
        spine_sql_query: str | None = None,
        resources: ResourceRequests | None = None,
        run_asynchronously: bool = False,
        store_online: bool = False,
        store_offline: bool = False,
        num_shards: int | None = None,
        num_workers: int | None = None,
        completion_deadline: timedelta | None = None,
        max_retries: int | None = None,
    ) -> Dataset:
        """Runs an evaluation on a set of prompts.
        See https://docs.chalk.ai/docs/prompts#prompt-evaluation for more information.

        Parameters
        ----------
        dataset_name
            The name of the `Dataset` to use for the evaluation.
            Dataset names are unique for each environment.
            If 'dataset_name' is provided, then 'dataset_id' should not be provided.
            If 'dataset_name' is provided along with 'inputs', then it will be used to
            generate save a `Dataset` constructed from the list of features computed
            from the inputs.
        dataset_id
            The UUID of the `Dataset` to use for the evaluation.
            Dataset ids are unique for each environment.
            If 'dataset_id' is provided, then 'dataset_name' and 'revision_id' should not be provided.
        revision_id
            The unique id of the `DatasetRevision` to use for the evaluation.
            If a previously-created dataset did not have a name, you can look it
            up using its unique job id instead.
            If 'revision_id' is provided, then 'dataset_name' and 'dataset_id' should not be provided.
        reference_output
            The name of the feature to use as the reference output for the evaluation.
        evaluators
            The list of evaluation functions to use for the evaluation.
            See https://docs.chalk.ai/docs/prompts#prompt-evaluation for more information.
        prompts
            The list of prompts to use for the evaluation.
            This can be a list of `Prompt` objects or a list of named prompts.
        meta
            Arbitrary `key:value` pairs to associate with a query.
        input
            The features for which there are known values.
            It can be a mapping of features to a list of values for each
            feature, or an existing `DataFrame`.
            Each element in the `DataFrame` or list of values represents
            an observation in line with the timestamp in `input_times`.
        spine_sql_query
            A SQL query that will query your offline store and use the result as input.
            See https://docs.chalk.ai/docs/query-offline#input for more information.
        input_times
            The time at which the given inputs should be observed for point-in-time correctness. If given a list of
            times, the list must match the length of the `input` lists. Each element of input_time corresponds with the
            feature values at the same index of the `input` lists.
            See https://docs.chalk.ai/docs/temporal-consistency for more information.
        output
            The features that you'd like to sample, if they exist.
            If an output feature was never computed for a sample (row) in
            the resulting `DataFrame`, its value will be `None`.
        recompute_features
            Used to control whether resolvers are allowed to run in order to compute feature values.

            If True, all output features will be recomputed by resolvers.
            If False, all output features will be sampled from the offline store.
            If a list, all output features in recompute_features will be recomputed,
            and all other output features will be sampled from the offline store.
        sample_features
            A list of features that will always be sampled, and thus always excluded from recompute.
            Should not overlap with any features used in `recompute_features` argument.
        environment
            The environment under which to run the resolvers.
            API tokens can be scoped to an environment.
            If no environment is specified in the query,
            but the token supports only a single environment,
            then that environment will be taken as the scope
            for executing the request.
        max_samples
            The maximum number of samples to include in the `DataFrame`.
            If not specified, all samples will be returned.
        branch
            If specified, Chalk will route your request to the relevant branch.
            If None, Chalk will route your request to a non-branch deployment.
            If not specified, Chalk will use the current client's branch info.
        correlation_id
            You can specify a correlation ID to be used in logs and web interfaces.
            This should be globally unique, i.e. a `uuid` or similar. Logs generated
            during the execution of your query will be tagged with this correlation id.
        query_context
            An immutable context that can be accessed from Python resolvers.
            This context wraps a JSON-compatible dictionary or JSON string with type restrictions.
            See https://docs.chalk.ai/api-docs#ChalkContext for more information.
        wait
            Whether to wait for job completion.
        show_progress
            If True, progress bars will be shown while the query is running.
            Primarily intended for use in a Jupyter-like notebook environment.
            This flag will also be propagated to the methods of the resulting
            `Dataset`.
        timeout:
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.
        lower_bound
            If specified, the query will only be run on data observed after this timestamp.
            Accepts strings in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
        upper_bound
            If specified, the query will only be run on data observed before this timestamp.
            Accepts strings in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
        store_plan_stages
            If True, the output of each of the query plan stages will be stored
            in S3/GCS. This will dramatically impact the performance of the query,
            so it should only be used for debugging.
            These files will be visible in the web dashboard's query detail view, and
            can be downloaded in full by clicking on a plan node in the query plan visualizer.
        tags
            The tags used to scope the resolvers.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        required_resolver_tags
            If specified, *all* required_resolver_tags must be present on a resolver for it to be
            considered eligible to execute.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        resources
            Override resource requests for processes with isolated resources, e.g., offline queries and cron jobs.
            See `ResourceRequests` for more information.
        run_asynchronously
            Boots a kubernetes job to run the queries in their own pods, separate from the engine and branch servers.
            This is useful for large datasets and jobs that require a long time to run.
        store_online
            If True, the output of the query will be stored in the online store.
        store_offline
            If True, the output of the query will be stored in the offline store.
        num_shards
            If specified, the query will be run asynchronously, splitting the input across `num_shards` shards.
        num_workers
            If specified, the query will be run asynchronously across a maximum `num_workers` pod workers at any time.
            This parameter is useful if you have a large number of shards and would like to limit the number of pods running at once.
        completion_deadline
            If specified, shards must complete within 'completion_deadline' duration, or they will be terminated.
            Terminated shards can be tried.
        max_retries
            If specified, failed offline query shards will be retried. The retry budget is shared across all shards.
            By default, max_retries=num_shards/

        Other Parameters
        ----------------
        required_output
            The features that you'd like to sample and must exist
            in each resulting row. Rows where a `required_output`
            was never stored in the offline store will be skipped.
            This differs from specifying the feature in `output`,
            where instead the row would be included, but the feature
            value would be `None`.

        Returns
        -------
        Dataset
            A Chalk `Dataset`.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> from chalk.prompts import Prompt, Message
        >>> dataset = ChalkClient().prompt_evaluation(
        ...     dataset_name='my_dataset',
        ...     reference_output='reference_output_column',
        ...     evaluators=['exact_match'],
        ...     prompts=[
        ...         Prompt(model='my_model', messages=[
        ...             Message(role='user', content='what is my name?'),
        ...         ]),
        ...     ]
        ... )
        >>> df = dataset.get_data_as_pandas()
        """
        ...

    async def upload_features(
        self,
        input: Mapping[FeatureReference, Any],
        branch: Optional[Union[BranchId, ellipsis]] = ...,
        environment: Optional[EnvironmentId] = None,
    ) -> UploadFeaturesResponse:
        ...

    async def get_operation_feature_statistics(self, operation_id: UUID) -> FeatureStatisticsResponse:
        """
        Fetches statistics for an operation's outputs.

        """
        ...

    async def __aenter__(self) -> AsyncChalkClient:
        ...

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exception: BaseException | None, tb: types.TracebackType | None
    ) -> None:
        ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.flag_pb2 import (
    GetFeatureFlagsRequest,
    GetFeatureFlagsResponse,
    SetFeatureFlagRequest,
    SetFeatureFlagResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class FeatureFlagServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    GetFeatureFlags: UnaryUnaryMultiCallable[
        GetFeatureFlagsRequest,
        GetFeatureFlagsResponse,
    ]
    SetFeatureFlag: UnaryUnaryMultiCallable[
        SetFeatureFlagRequest,
        SetFeatureFlagResponse,
    ]

class FeatureFlagServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def GetFeatureFlags(
        self,
        request: GetFeatureFlagsRequest,
        context: ServicerContext,
    ) -> GetFeatureFlagsResponse: ...
    @abstractmethod
    def SetFeatureFlag(
        self,
        request: SetFeatureFlagRequest,
        context: ServicerContext,
    ) -> SetFeatureFlagResponse: ...

def add_FeatureFlagServiceServicer_to_server(servicer: FeatureFlagServiceServicer, server: Server) -> None: ...



from chalk._gen.chalk.common.v1 import chalk_error_pb2 as _chalk_error_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class UploadFeaturesRequest(_message.Message):
    __slots__ = ("inputs_table",)
    INPUTS_TABLE_FIELD_NUMBER: _ClassVar[int]
    inputs_table: bytes
    def __init__(self, inputs_table: _Optional[bytes] = ...) -> None: ...

class UploadFeaturesResponse(_message.Message):
    __slots__ = ("errors", "operation_id")
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    OPERATION_ID_FIELD_NUMBER: _ClassVar[int]
    errors: _containers.RepeatedCompositeFieldContainer[_chalk_error_pb2.ChalkError]
    operation_id: str
    def __init__(
        self,
        errors: _Optional[_Iterable[_Union[_chalk_error_pb2.ChalkError, _Mapping]]] = ...,
        operation_id: _Optional[str] = ...,
    ) -> None: ...
